from Aspidites._vendor.contracts.test_registrar import (good_examples, semantic_fail_examples,
                                                        contract_fail_examples)
from Aspidites._vendor.contracts import parse, ContractSyntaxError


def get_all_strings():
    all_strings = (good_examples + semantic_fail_examples +
                   contract_fail_examples)
    for contract, _, _ in all_strings:
        if isinstance(contract, list):
            for c in contract:
                yield c
        else:
            yield contract

# went from 89.9% to 95.7%


def main():
    examples = get_all_strings()

    differences = run_joker(examples)

    diff = list(differences)
    unfriendliness = sum(diff) / len(diff)

    friendliness = 100 - 100 * unfriendliness
    print("Friendliness: %.2f%% " % friendliness)


def replace_one(s, i, c):
    assert i >= 0 and i < len(s)
    return s[:i] + c + s[i + 1:]

assert replace_one('python', 1, 'a') == 'pathon'
s = 'python'
for i in range(len(s)):
    s2 = replace_one(s, i, '~')
    s3 = replace_one(s2, i, s[i])
    assert s == s3, 'i=%d  %r -> %r -> %r' % (i, s, s2, s3)


def run_joker(examples):

    for s in examples:
        # make sure we can parse it
        parse(s)

        # now alter one letter
        for i in range(len(s)):
            s2 = replace_one(s, i, '~')

            try:
                parse(s2)

            except ContractSyntaxError as e:
                detected = e.where.col - 1
                displacement = i - detected
#                if  displacement < 0:
#                    print displacement
#                    print e
                assert displacement >= 0

                value = displacement * 1.0 / len(s)
                if displacement > 0:
                    print(e)
#                    assert False

                yield value


if __name__ == '__main__':
    main()



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.
import inspect
import operator as op
import shutil
import sys
import traceback
import typing
from math import inf, isinf
from contextlib import suppress

from _warnings import warn
from pyrsistent import v
from Aspidites import final
from Aspidites._vendor.contracts import ContractNotRespected, contract
from Aspidites._vendor.fn import apply
from Aspidites._vendor.fn.underscore import ArityError

from Aspidites.api import ContractBreachWarning, create_warning


# noinspection PyPep8Naming
def SafeDiv(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0:
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeDiv, a=a, b=b)
    return a / b


# noinspection PyPep8Naming
def SafeMod(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0:
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Modulus by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeMod, a=a, b=b)
    return a % b


# noinspection PyPep8Naming
def SafeExp(a, b):
    if (a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0):  # 0**0, inf**0, 0**inf
        stack = inspect.stack()
        w = create_warning(
            stack[0][3], [a, b], {}, stack,
            ArithmeticError("%s**%s" % (str(a), str(b),) + " == Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeExp, a=a, b=b)
    return a**b


class Maybe:
    """Sandboxes a Surely call and handles ContractNotRespected by returning Undefined"""

    __metaclass__ = final
    __slots__ = v("_func", "_args", "_kwargs", "__instance__")

    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.__instance__ = Undefined()

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + ", ".join(
                [
                    str(self._func.__name__),
                    str(self._args).strip("()"),
                    *[str(k) + " = " + str(v) for k, v in self._kwargs.items()],
                ]
            )
            + ")"
            + (
                " -> %s" % Undefined()
                if self.__instance__ == Undefined()
                else " -> %s" % str(self.__instance__)
            )
        )

    def __invert__(self):
        return ~self.__instance__

    def __neg__(self):
        return -self.__instance__

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def __call__(self, debug=False, warn_undefined=True):
        try:
            with suppress(ValueError):
                val = apply(self.func, self.args, self.kwargs)
            with suppress(UnboundLocalError):
                self.__instance__ = Surely(val)
                # SURELY #
                return self.__instance__
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
        except (ContractNotRespected, ArityError, ZeroDivisionError) as e:
            if warn_undefined:
                scope = len(inspect.trace()) if debug else 1
                for i in range(scope):
                    stack = inspect.stack(scope)
                    w = create_warning(self.func, self.args, self.kwargs, stack, e)
                    warn(
                        w,
                        category=ContractBreachWarning
                        if isinstance(e, ContractNotRespected)
                        else RuntimeWarning,
                    )
            # UNDEFINED #
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
            return self.__instance__


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton"""

    __metaclass__ = final
    __slots__ = v("__weakref__", "__instance__")
    __instance = None

    def __hash__(self):
        return hash(self.__instance__)

    def __eq__(self, other):
        return self.__hash__ == other.__hash__

    def __add__(self, other):
        return other

    def __sub__(self, other):
        return -other

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __str__(self):
        return str(self.__repr__())

    def __float__(self):
        return float()

    def __complex__(self):
        return complex()

    def __oct__(self):
        return oct(0)

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        return self.__class__.__name__

    def __nonzero__(self):
        return False

    def __call__(self, *args, **kwargs):
        return self.__new__(self.__class__, *args, **kwargs)

    def __new__(mcs, *args, **kwargs):
        if mcs.__instance is None:
            mcs.__instance = super(Undefined, mcs).__new__(mcs, *args, **kwargs)
            mcs.__instance__ = mcs.__instance
        return mcs.__instance__  # instance descriptor from __slots__ -> actual instance


class Surely:
    """A monad for a successful programmatic unit
    Truthy, defers to an instance of a successful computation"""

    __metaclass__ = final
    __slots__ = v(
        "__weakref__", "__instance__" "__str__", "__int__", "__float__", "__complex__"
    )

    def __try_except_undefined__(self, other=None, call=None):
        if call:
            # noinspection PyComparisonWithNone
            return (
                call(self.__instance__)
                if other == None
                else call(self.__instance__, other)
            )
        else:
            return self.__instance__

    def __hash__(self):
        return self.__try_except_undefined__(call=hash)

    def __eq__(self, other):
        return self.__try_except_undefined__(other, op.eq)

    def __add__(self, other):
        return self.__try_except_undefined__(other, op.add)

    def __sub__(self, other):
        return self.__try_except_undefined__(other, op.sub)

    def __neg__(self):
        return self.__try_except_undefined__(call=op.neg)

    def __invert__(self):
        return self.__try_except_undefined__(call=op.invert)

    def __mul__(self, other):
        return self.__try_except_undefined__(other, op.mul)

    def __truediv__(self, other):
        return self.__try_except_undefined__(other, op.truediv)

    def __floordiv__(self, other):
        return self.__try_except_undefined__(other, op.floordiv)

    def __oct__(self):
        return self.__try_except_undefined__(call=oct)

    def __nonzero__(self):
        return self.__try_except_undefined__(call=bool)

    @classmethod
    def __call__(cls, *args, **kwargs):
        return cls.__instance__

    def __new__(cls, instance__=Undefined(), *args, **kwargs):
        cls.__instance__ = instance__
        return cls.__instance__



from Aspidites.woma import *
from Aspidites._vendor import F as F
from Aspidites._vendor.RestrictedPython import safe_builtins as safe_builtins
from Aspidites._vendor.contracts import contract as contract, new_contract as new_contract
from Aspidites.monads import Maybe as Maybe, SafeDiv as SafeDiv, SafeExp as SafeExp, SafeMod as SafeMod, Surely as Surely, Undefined as Undefined
from collections.abc import Generator
from pyrsistent import PClass as PClass, PRecord as PRecord, m as m, s as s, v as v
from typing import Any

procedure: None
coroutine: Generator
number: Any

def Add(x: int = ..., y: int = ...) -> int: ...
def Yield123() -> coroutine: ...
def Hello() -> procedure: ...
def Hello2() -> coroutine: ...

D: Any
G: Any
C: Any
x: Any
y: Any
z: Any
scala: Any
val: Any
denom: int
div_by_zero: Any
mod_zero: Any
div_by_zero2: Any
mod_zero2: Any



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.

import os
import py_compile
import sys
from glob import glob
from warnings import warn

from mypy import api

from Aspidites._vendor.contracts import contract
from Aspidites.api import CheckedFileStack, checksum
from Aspidites.templates import lib, makefile, pyproject, setup


@contract()
def compile_module(
    code: "code",
    fname: "str" = "compiled.py",
    force: "bool" = False,
    bytecode: "bool" = False,
    c: "bool" = True,
    build_requires: "list|str" = "",
    verbose: "int" = 0,
    *args,
    **kwargs
):

    app_name = os.path.splitext(fname)[0]
    project = os.path.basename(app_name)
    module_name = app_name.replace("/", ".")
    file_c = app_name + ".c"
    dir = os.path.dirname(file_c)
    glob_so = app_name + ".*.so"
    init_py = os.path.join(dir, "__init__.py")
    make_ = os.path.join(dir, "Makefile")
    py_typed = os.path.join(dir, "py.typed")
    stack = CheckedFileStack()
    mode = "x" if force else "w"
    open(fname, mode).write(lib.substitute(code="\n".join(code)))
    stack.register(fname)
    open(py_typed, "w").write("# THIS FILE IS GENERATED - DO NOT EDIT #")
    stack.register(py_typed)
    open(init_py, "w").write("# THIS FILE IS GENERATED - DO NOT EDIT #")
    stack.register(init_py)
    open(make_, mode).write(makefile.substitute(project=project))
    stack.register(make_)
    verb = int(bool(verbose))
    mypy_args = [
        "-m",
        module_name,
        "--follow-imports=skip",
        "--show-error-context",
        "--show-error-codes",
        "--allow-incomplete-defs",
        "--disable-error-code=valid-type",
        # '--disallow-untyped-defs',
        # '--disallow-untyped-calls',
    ]
    print("running mypy", " ".join(mypy_args))
    type_report = None
    error_report = None
    return_code = 0
    try:
        type_report, error_report, return_code = api.run(mypy_args)
    except AttributeError:
        warn("mypy api call failed")

    finally:
        print("mypy type report: ", type_report) if type_report else None
        print("mypy error report: ", error_report) if error_report else None
        print("mypy returned with exit code:", return_code) if return_code else None
        # exit(return_code) if return_code != 0 else print("running compile")

    fname_pyi = app_name + '.pyi'
    stubgen_runner = 'stubgen %s -o .' % (fname)
    print("running %s" % stubgen_runner)
    with os.popen(stubgen_runner) as p:
        print(p.read())
    try:
        stack.register(fname_pyi)
    except FileNotFoundError as e:
        warn(str(e))
        print("running rename __main__.pyi to %s" % fname_pyi)
        os.rename('__main__.pyi', fname_pyi)
        stack.register(fname_pyi)

    if bytecode:
        fname_pyc = app_name + ".pyc"
        quiet = tuple(reversed(range(3))).index(verbose if verbose < 2 else 2)
        py_compile.compile(fname, fname_pyc, quiet=quiet)
        stack.register(fname_pyc)

    if c:
        os.popen("cython %s %s %s" % (fname, "--force" * force, "--verbose" * verb))
        setup_py = os.path.join(dir, "setup.py")
        pyproject_toml = os.path.join(dir, "pyproject.toml")
        with open(setup_py, mode) as f:
            f.write(
                setup.substitute(
                    app_name=module_name,
                    src_file=fname,
                    inc_dirs=[],
                    libs=[],
                    exe_name=app_name,
                    lib_dirs=[],
                    **kwargs
                )
            )
        stack.register(setup_py)
        setup_runner = "%s %s build_ext -b ." % (sys.executable, setup_py)
        print("running", setup_runner)
        with os.popen(setup_runner) as p:
            print(p.read())
        stack.register(file_c)
        for i in glob(glob_so):
            stack.register(i)
        with open(pyproject_toml, mode) as f:
            f.write(pyproject.substitute(build_requires=build_requires))
        stack.register(pyproject_toml)
    all_file_checksums = stack.finalize()
    print("running checksums")
    for k, v in all_file_checksums.items():
        digest = checksum(v, write=False, check=True)
        try:
            all_file_checksums.get(digest)
        except AttributeError:  # 449779cbdc60682faf8b1327d1d315ca
            raise RuntimeError("\nfor file %s\n%s\n  did not match cached digest\n%s")



from .utils import check_contracts_fail
from Aspidites._vendor.contracts import ContractNotRespected, parse, Contract
from Aspidites._vendor.contracts.test_registrar import (semantic_fail_examples,
                                                        contract_fail_examples, good_examples)
import pickle


def check_exception_pickable(contract, value):
    exception = check_contracts_fail(contract, value)
    assert isinstance(exception, Exception)
    try:
        s = pickle.dumps(exception)
        pickle.loads(s)
    except TypeError as e:
        print('While pickling: %s' % exception)
        raise e
        # msg = 'Could not pickle exception.\n'
        # msg += str(exception)
        # msg += 'Raised: %s' % e
        # raise Exception(msg)


# def test_exceptions_are_pickable():
#     for contract, value, exact in semantic_fail_examples:  # @UnusedVariable
#         yield check_contracts_fail, contract, value, ContractNotRespected
#         #ContractSemanticError
#     for contract, value, exact in contract_fail_examples:  # @UnusedVariable
#         yield check_contracts_fail, contract, value, ContractNotRespected


def check_contract_pickable(contract):
    c = parse(contract)
    assert isinstance(c, Contract)
    try:
        s = pickle.dumps(c)
        c2 = pickle.loads(s)
    except TypeError as e:
        msg = 'Could not pickle contract.\n'
        msg += '- string: %s\n' % c
        msg += '-   repr: %r\n' % c
        msg += 'Exception: %s' % e
        raise Exception(msg)

    assert c == c2


# def test_contracts_are_pickable():
#     allc = (good_examples + semantic_fail_examples + contract_fail_examples)
#     for contract, _, _ in allc:
#         if isinstance(contract, list):
#             for c in contract:
#                 yield check_contract_pickable, c
#         else:
#             yield check_contract_pickable, contract



from Aspidites.woma import *
from Aspidites._vendor import F as F
from Aspidites._vendor.contracts import new_contract as new_contract
from Aspidites.monads import Surely as Surely, Undefined as Undefined
from collections.abc import Generator
from pyrsistent import PClass as PClass, PRecord as PRecord, m as m, pmap as pmap, pset as pset, pvector as pvector, s as s, v as v
from typing import Any

procedure: None
coroutine: Generator
number: Any

def Add(x: number = ..., y: number = ...) -> number: ...
def Sub(x: number = ..., y: number = ...) -> number: ...
def Div(x: number = ..., y: number = ...) -> number: ...
def Exp(x: number = ..., y: number = ...) -> number: ...
def Mod(x: number = ..., y: number = ...) -> number: ...



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.
from string import Template

makefile = Template("""clean: clean-build clean-pyc clean-sha256 ## remove all build, test, coverage and Python artifacts

project = $project

uninstall: distclean
	rm -fr py.typed
	rm -fr setup.py
	rm -fr pyproject.toml
	find . -name '$$(project)*' -not -name '*.wom' -exec rm -fr {} +

distclean: clean-build clean-pyc clean-sha256
	rm -fr Makefile

clean-build: ## remove build artifacts
	rm -fr build/
	rm -fr dist/
	rm -fr .eggs/
	find . -name '*.egg-info' -exec rm -fr {} +
	find . -name '*.egg' -exec rm -f {} +

clean-sha256:
	find . -name '*.sha256' -exec rm -f {} +

clean-pyc: ## remove Python file artifacts
	find . -name '*.pyc' -exec rm -f {} +
	find . -name '*.pyo' -exec rm -f {} +
	find . -name '*~' -exec rm -f {} +
	find . -name '__pycache__' -exec rm -fr {} +
""")

_warning = Template("""
$file:$lineno:
$func
Statement:
$atfault
Bound Variables:
$bound
Complaint Information:
$tb
""")

setup = Template("""
# THIS FILE IS GENERATED - DO NOT EDIT #
from setuptools import setup, Extension
from Cython.Build import cythonize
from Cython.Compiler import Options
Options.annotate = $annotate
Options.annotate_coverage_xml = $annotate_coverage_xml
Options.buffer_max_dims = $buffer_max_dims
Options.cache_builtins = $cache_builtins
Options.cimport_from_pyx = $cimport_from_pyx
Options.clear_to_none = $clear_to_none
Options.closure_freelist_size = $closure_freelist_size
Options.convert_range = $convert_range
Options.docstrings = $docstrings
Options.embed_pos_in_docstring = $embed_pos_in_docstring
Options.generate_cleanup_code = $generate_cleanup_code
Options.fast_fail = $fast_fail
Options.warning_errors = $warning_errors
Options.error_on_unknown_names = $error_on_unknown_names
Options.error_on_uninitialized = $error_on_uninitialized
Options.gcc_branch_hints = $gcc_branch_hints
Options.lookup_module_cpdef = $lookup_module_cpdef
Options.embed = $embed

exts = [
Extension(
         '$app_name', 
         ['$src_file'],
         include_dirs=$inc_dirs,
         libraries=$libs, 
         extra_compile_args=['-Wall'],
         library_dirs=$lib_dirs
    ),
]

setup(
    name='$app_name',
    ext_modules=cythonize(exts))
""")

lib = Template("""# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8
# THIS FILE IS GENERATED - DO NOT EDIT #
import cython  # type: ignore
from typing import Any
from collections.abc import Generator
from pyrsistent import (
    pset, 
    pmap, 
    pvector, 
    s, v, m, 
    PRecord,
    PClass
)
from Aspidites.woma import *
from Aspidites._vendor import F, _
from Aspidites.monads import Maybe, Surely, Undefined, SafeDiv, SafeMod, SafeExp
from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.RestrictedPython import safe_builtins
# DECLARATIONS TO ALLOW CONTRACTS TO TYPE CHECK #
procedure: None
coroutine: Generator
number: Any
globals().update(dict(__builtins__=safe_builtins))  # add all imports to globals


$code

""")

pyproject = Template("""[build-system]
requires = ["setuptools", "wheel", "Cython", $build_requires]""")



import unittest

from Aspidites._vendor import match_dict, match
from Aspidites._vendor import ANY


class IterableTests(unittest.TestCase):

    def test_match_dict(self):
        self.assertEqual(match_dict({'a': ANY, 'b': 2}, {'a': 1, 'b': 2}), (True, [1]))

    def test_match_dict_ordering(self):
        for i in range(100):
            self.assertEqual(match_dict({'a': ANY, 'b': ANY}, {'a': 1, 'b': 2}), (True, [1, 2]))

    def test_match_asymmetric(self):
        self.assertEqual(match_dict({'a': ANY, 'b': 2},           {'a': 1, 'b': 2, 'c': 3}),    (True, [1]))
        self.assertEqual(match_dict({'a': ANY, 'b': 2, 'c': 3},   {'a': 1, 'b': 2}),            (False, []))

    def test_match_nested(self):
        self.assertEqual(match_dict({'a': {'b': ANY}, 'c': ANY},
                                    {'a': {'b': 1}, 'c': 2}), (True, [1, 2]))

    def test_dog(self):
        pet = {'type': 'dog', 'details': {'age': 3}}

        self.assertEqual(match(pet, {'details': {'age': ANY}},    lambda age: age),       3)
        self.assertEqual(match(pet, {        ANY: {'age': ANY}},    lambda a, b: (a, b)),   ('details', 3))

    def test_exclude_previously_used_keys(self):
        x = {"a": 1, "b": 2}
        self.assertEqual(match_dict({"a": ANY, ANY: ANY  }, x), (True, [1, "b", 2]))
        self.assertEqual(match_dict({"a": ANY, ANY: int}, x), (True, [1, "b", 2]))
        self.assertEqual(match_dict({"a": ANY, ANY: int}, x), (True, [1, "b", 2]))

    def test_multi_underscore_ambiguous(self):
        for i in range(20):
            self.assertEqual(match_dict({"a": ANY, ANY: int},
                                        {"a": 1, "b": 2, "c": 3}), (True, [1, "b", 2]))

            self.assertEqual(match_dict({"type": ANY, ANY: str},
                                        {"type": "pet", "cat-name": "bonney", "info": {"age": 1}}),
                             (True, ["pet", "cat-name", "bonney"]))

    def test_wild_dicts(self):
        data = [
            {"type": "dog", "dog-name": "fuffy", "info": {"age": 2}},
            {"type": "pet", "dog-name": "puffy", "info": {"age": 1}},
            {"type": "cat", "cat-name": "buffy", "cat-info": {"age": 3}},
        ]

        # I want the average age of a pet, but the data is inconsistent :/
        ages = [match(row, {ANY: {"age": int}}, lambda field, age: age) for row in data]
        average_age = sum(ages) / len(ages)
        self.assertEqual(average_age, (2+1+3)/3)

        # I want al the names, but data is inconsistent!
        names = [match(row, {"type": ANY, ANY: str}, lambda type, name_field, name: name) for row in data]
        self.assertEqual(names, ['fuffy', 'puffy', 'buffy'])



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor._compat import IS_PY3
from Aspidites._vendor.RestrictedPython.Guards import guarded_unpack_sequence
from Aspidites.tests.helper import restricted_exec

import pytest


functiondef_err_msg = 'Line 1: "_bad" is an invalid variable ' \
                      'name because it starts with "_"'


def test_RestrictingNodeTransformer__visit_FunctionDef__1():
    """It prevents function arguments starting with `_`."""
    result = compile_restricted_exec("def foo(_bad): pass")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and **_bad
    # would be allowed.
    assert functiondef_err_msg in result.errors


def test_RestrictingNodeTransformer__visit_FunctionDef__2():
    """It prevents function keyword arguments starting with `_`."""
    result = compile_restricted_exec("def foo(_bad=1): pass")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and **_bad
    # would be allowed.
    assert functiondef_err_msg in result.errors


def test_RestrictingNodeTransformer__visit_FunctionDef__3():
    """It prevents function * arguments starting with `_`."""
    result = compile_restricted_exec("def foo(*_bad): pass")
    assert result.errors == (functiondef_err_msg,)


def test_RestrictingNodeTransformer__visit_FunctionDef__4():
    """It prevents function ** arguments starting with `_`."""
    result = compile_restricted_exec("def foo(**_bad): pass")
    assert result.errors == (functiondef_err_msg,)


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in Python 3")
def test_RestrictingNodeTransformer__visit_FunctionDef__5():  # pragma: PY2
    """It prevents function arguments starting with `_` in tuples."""
    result = compile_restricted_exec("def foo((a, _bad)): pass")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and **_bad
    # would be allowed.
    assert functiondef_err_msg in result.errors


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in Python 3")
def test_RestrictingNodeTransformer__visit_FunctionDef__6():  # pragma: PY2
    """It prevents function arguments starting with `_` in tuples."""
    result = compile_restricted_exec("def foo(a, (c, (_bad, c))): pass")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and
    # **_bad would be allowed.
    assert functiondef_err_msg in result.errors


@pytest.mark.skipif(
    IS_PY2,
    reason="There is no single `*` argument in Python 2")
def test_RestrictingNodeTransformer__visit_FunctionDef__7():  # pragma: PY3
    """It prevents `_` function arguments together with a single `*`."""
    result = compile_restricted_exec("def foo(good, *, _bad): pass")
    assert result.errors == (functiondef_err_msg,)


NESTED_SEQ_UNPACK = """
def nested((a, b, (c, (d, e)))):
    return a, b, c, d, e

def nested_with_order((a, b), (c, d)):
    return a, b, c, d
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in python 3")
def test_RestrictingNodeTransformer__visit_FunctionDef__8(
        mocker):  # pragma: PY2
    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda it: it

    glb = {
        '_getiter_': _getiter_,
        '_unpack_sequence_': guarded_unpack_sequence
    }

    restricted_exec('def simple((a, b)): return a, b', glb)

    val = (1, 2)
    ret = glb['simple'](val)
    assert ret == val
    _getiter_.assert_called_once_with(val)
    _getiter_.reset_mock()

    restricted_exec(NESTED_SEQ_UNPACK, glb)

    val = (1, 2, (3, (4, 5)))
    ret = glb['nested'](val)
    assert ret == (1, 2, 3, 4, 5)
    assert 3 == _getiter_.call_count
    _getiter_.assert_any_call(val)
    _getiter_.assert_any_call(val[2])
    _getiter_.assert_any_call(val[2][1])
    _getiter_.reset_mock()

    ret = glb['nested_with_order']((1, 2), (3, 4))
    assert ret == (1, 2, 3, 4)
    _getiter_.assert_has_calls([
        mocker.call((1, 2)),
        mocker.call((3, 4))])
    _getiter_.reset_mock()


BLACKLISTED_FUNC_NAMES_CALL_TEST = """
def __init__(test):
    test

__init__(1)
"""


def test_RestrictingNodeTransformer__module_func_def_name_call():
    """It forbids definition and usage of magic methods as functions ...

    ... at module level.
    """
    result = compile_restricted_exec(BLACKLISTED_FUNC_NAMES_CALL_TEST)
    # assert result.errors == ('Line 1: ')
    assert result.errors == (
        'Line 2: "__init__" is an invalid variable name because it starts with "_"',  # NOQA: E501
        'Line 5: "__init__" is an invalid variable name because it starts with "_"',  # NOQA: E501
    )



import unittest

from Aspidites._vendor.contracts.docstring_parsing import DocStringInfo, Arg, number_of_spaces
from Aspidites._vendor.contracts.interface import add_prefix


examples = {"""
        Provides a RGB representation of the values by interpolating the range
        [min(value),max(value)] into the colorspace [min_color, max_color].

        :param value: The field to represent.
        :type  value: HxW array

        :param max_value: If specified, everything *above* is clipped.
        :type max_value: float
        :param min_value: If specified, everything *below* is clipped.
        :type min_value: float

        :param min_color:  Color to give to the minimum values.


        :return: A RGB image.
        :rtype: HxWx3 uint8

        :return: gray

    """: DocStringInfo(docstring='\n        Provides a RGB representation of the values by interpolating the range\n'
                         '        [min(value),max(value)] into the colorspace [min_color, max_color].\n',
  params={
        'value': Arg('The field to represent.', 'HxW array'),
        'max_value': Arg('If specified, everything *above* is clipped.', 'float'),
        'min_value': Arg('If specified, everything *below* is clipped.', 'float'),
        'min_color': Arg('Color to give to the minimum values.', None),
  },
  returns=[Arg('A RGB image.', "HxWx3 uint8"), Arg('gray', None)]
)
}


class DocStringTest(unittest.TestCase):

    def test_parsing(self):
        for string in examples:
            parsed = DocStringInfo.parse(string)
            "%s" % parsed
            "%r" % parsed
            result = examples[string]
            self.assertEqual(result, parsed)

    def test_number_of_spaces(self):
        self.assertEqual(number_of_spaces(''), 0)
        self.assertEqual(number_of_spaces(' '), 1)
        self.assertEqual(number_of_spaces('  '), 2)
        self.assertEqual(number_of_spaces('11'), 0)
        self.assertEqual(number_of_spaces(' 223'), 1)
        self.assertEqual(number_of_spaces('  4343'), 2)

    def test_reparsing(self):
        for string, result in examples.items(): #@UnusedVariable
            parsed = DocStringInfo.parse(string)
            converted = "%s" % parsed
            reparsed = DocStringInfo.parse(converted)

            msg = ('First string:\n%s\nParsed as:\n%s\n' %
                (add_prefix(string, '|'), add_prefix('%r' % parsed, '|')))

            msg += ('Converted:\n%s\nReparsed as:\n%s\n' %
                (add_prefix(converted, '|'), add_prefix('%r' % reparsed, '|')))

            self.assertEqual(parsed, reparsed, msg=msg)

    def test_inline_params(self):
        def test_inline_parsing(docstring, expected_type="type",
                                expected_desc="desc"):
            info = DocStringInfo.parse(docstring)
            self.assertTrue("name" in info.params)
            self.assertEqual(info.params["name"].type, expected_type)
            self.assertEqual(info.params["name"].desc, expected_desc)

        # Proper syntax
        test_inline_parsing(":param type name: desc")
        test_inline_parsing(":param name: desc", None)
        test_inline_parsing(":param name:", None, None)

        # Weird syntax for people who like to break things.
        test_inline_parsing(" : param type name : desc ")
        test_inline_parsing(" : param  name : desc ", None)
        test_inline_parsing(" : param  name : ", None, None)
        test_inline_parsing(" : param type , > 0  name : ", "type , > 0", None)

    def test_inline_returns(self):
        def test_inline_parsing(docstring, expected_type="type",
                                expected_desc="desc"):
            info = DocStringInfo.parse(docstring)
            self.assertTrue(len(info.returns) > 0)
            self.assertEqual(info.returns[0].type, expected_type)
            self.assertEqual(info.returns[0].desc, expected_desc)

        # Proper syntax
        test_inline_parsing(":returns type: desc")
        test_inline_parsing(":returns: desc", None)
        test_inline_parsing(":returns:", None, None)

        # Weird syntax for people who like to break things.
        test_inline_parsing(" : returns type : desc ")
        test_inline_parsing(" : returns : desc ", None)
        test_inline_parsing(" : returns : ", None, None)
        test_inline_parsing(" : returns type , > 0 : ", "type , > 0", None)



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.

import contextlib
import os
import typing
from hashlib import sha256
from inspect import isfunction, signature
from os import get_terminal_size
from textwrap import wrap as _wrap

from pyparsing import ParseResults
from pyrsistent import PVector, discard, inc, pmap, pvector, rex

from ._vendor.contracts import new_contract
from ._vendor.fn import F, _
from ._vendor.fn.underscore import ArityError, _Callable
from .templates import _warning


class ContractBreachWarning(RuntimeWarning):
    pass


def wrap(text, width=160, pad=True, padchar=" "):
    """
    Do not remove whitespaces in string but still wrap text to max width.
    Instead of passing the entire text to textwrap.wrap, split and pass each
    line instead. This way list formatting is not mangled by textwrap.wrap.
    """
    wrapped_lines = []
    try:
        for l in text.splitlines():
            line = _wrap(l, width, replace_whitespace=False)
            if pad:
                for s in line:
                    s += padchar * width
            wrapped_lines.extend(line)
    except ArityError:
        wrapped_lines.extend([str(text)])

    return wrapped_lines


def bordered(text, width=160):
    lines = [i for i in wrap(text, width=width)]
    width = max((len(s) for s in lines), default=width) or width
    res = ["╭" + "┉" * width + "╮"]
    for s in lines:
        while len(s) < width:
            s += " "
        res.append("┊" + s + "┊")
    res.append("╰" + "┉" * width + "╯")
    return "\n".join(res)


def format_kwargs(kwargs: "dict", sep=", "):
    return sep + str(kwargs).strip("{} ").replace(":", "=") if len(kwargs) else ""


def format_locals(locals, exc: "Exception"):
    locals_ = dict(filter(lambda x: x[1] != str(exc), locals))
    str_locals = str()
    for k, v_ in locals_.items():
        if str(k).startswith("@"):  # skip @py_assert
            continue
        if isfunction(v_):
            str_locals += k + ": " + str(signature(v_)).replace("'", "") + "\n"
        else:
            str_locals += k + ": " + str(v_) + "\n"
    return str_locals.rstrip("\n")


def create_warning(func, args, kwargs, stack, exc=Exception()):
    _locals = stack[1][0].f_locals.items()
    str_locals = format_locals(_locals, exc)
    func_name = stack[1][0].f_code.co_name
    fname = stack[1][0].f_code.co_filename
    lineno = stack[1][0].f_code.co_firstlineno
    fkwargs = format_kwargs(kwargs)
    if hasattr(func, "__name__"):
        name = func.__name__
    else:
        name = str(func)
    atfault = (
        name
        if isinstance(exc, ArityError)
        else name + "(" + str(args).strip("()") + fkwargs + ")"
    )
    return _warning.safe_substitute(
        file=fname,
        lineno=lineno,
        func=bordered(func_name),
        atfault=bordered(atfault),
        bound=bordered(str_locals),
        tb=bordered(str(exc)),
    )


MD5 = ".sha256"
code = new_contract("code", lambda x: isinstance(x, ParseResults))


@contextlib.contextmanager
def working_directory(path):
    """
    A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.
    Usage:
    > # Do something in original directory
    > with working_directory('/my/new/path'):
    >     # Do something in new directory
    > # Back to old directory
    """

    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def checksum(fname, write=True, check=False):
    base, name = os.path.split(fname)
    fname_sha256 = os.path.join(base, "." + name) + ".sha256"

    def read_sha256(data):
        curr_hash = sha256()
        chunk = data.read(8192)
        while chunk:
            curr_hash.update(chunk)
            chunk = data.read(8192)
        return curr_hash

    if write:
        with open(fname, "rb") as data:
            curr_hash = read_sha256(data)
            with open(fname_sha256, "wb") as digest:
                digest.write(curr_hash.digest())
            return pmap({curr_hash.digest(): fname}).items()[0]  # immutable
    if check:
        with open(fname_sha256, "rb") as digest:
            with open(fname, "rb") as data:
                curr_hash = read_sha256(data)
                old = digest.read()
                new = curr_hash.digest()
                if new == old:
                    print(
                        "sha256 digest check successful: %s, %s == %s"
                        % (fname, new.hex(), old.hex())
                    )
                    return new
                else:
                    print(
                        "sha256 digest failure: %s, %s != %s"
                        % (fname, new.hex(), old.hex())
                    )
                    return ""


class CheckedFileStack:
    def __init__(self, initial=None, pre_size=8192):
        if initial is None:
            initial = {}
        self._files = pmap(initial, pre_size)
        self.all_files = self._files.evolver()

    def register(self, fname):
        self.all_files.set(*checksum(fname))

    def finalize(self):
        return self.all_files.persistent()



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor.RestrictedPython.Guards import guarded_unpack_sequence
from Aspidites.tests.helper import restricted_exec

import contextlib


WITH_STMT_WITH_UNPACK_SEQUENCE = """
def call(ctx):
    with ctx() as (a, (c, b)):
        return a, c, b
"""


def test_with_stmt_unpack_sequence(mocker):
    @contextlib.contextmanager
    def ctx():
        yield (1, (2, 3))

    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda ob: ob

    glb = {
        '_getiter_': _getiter_,
        '_unpack_sequence_': guarded_unpack_sequence
    }

    restricted_exec(WITH_STMT_WITH_UNPACK_SEQUENCE, glb)

    ret = glb['call'](ctx)

    assert ret == (1, 2, 3)
    _getiter_.assert_has_calls([
        mocker.call((1, (2, 3))),
        mocker.call((2, 3))])


WITH_STMT_MULTI_CTX_WITH_UNPACK_SEQUENCE = """
def call(ctx1, ctx2):
    with ctx1() as (a, (b, c)), ctx2() as ((x, z), (s, h)):
        return a, b, c, x, z, s, h
"""


def test_with_stmt_multi_ctx_unpack_sequence(mocker):
    result = compile_restricted_exec(WITH_STMT_MULTI_CTX_WITH_UNPACK_SEQUENCE)
    assert result.errors == ()

    @contextlib.contextmanager
    def ctx1():
        yield (1, (2, 3))

    @contextlib.contextmanager
    def ctx2():
        yield (4, 5), (6, 7)

    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda ob: ob

    glb = {
        '_getiter_': _getiter_,
        '_unpack_sequence_': guarded_unpack_sequence
    }

    exec(result.code, glb)

    ret = glb['call'](ctx1, ctx2)

    assert ret == (1, 2, 3, 4, 5, 6, 7)
    _getiter_.assert_has_calls([
        mocker.call((1, (2, 3))),
        mocker.call((2, 3)),
        mocker.call(((4, 5), (6, 7))),
        mocker.call((4, 5)),
        mocker.call((6, 7))
    ])


WITH_STMT_ATTRIBUTE_ACCESS = """
def simple(ctx):
    with ctx as x:
        x.z = x.y + 1

def assign_attr(ctx, x):
    with ctx as x.y:
        x.z = 1

def load_attr(w):
    with w.ctx as x:
        x.z = 1

"""


def test_with_stmt_attribute_access(mocker):
    _getattr_ = mocker.stub()
    _getattr_.side_effect = getattr

    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob

    glb = {'_getattr_': _getattr_, '_write_': _write_}
    restricted_exec(WITH_STMT_ATTRIBUTE_ACCESS, glb)

    # Test simple
    ctx = mocker.MagicMock(y=1)
    ctx.__enter__.return_value = ctx

    glb['simple'](ctx)

    assert ctx.z == 2
    _write_.assert_called_once_with(ctx)
    _getattr_.assert_called_once_with(ctx, 'y')

    _write_.reset_mock()
    _getattr_.reset_mock()

    # Test assign_attr
    x = mocker.Mock()
    glb['assign_attr'](ctx, x)

    assert x.z == 1
    assert x.y == ctx
    _write_.assert_has_calls([
        mocker.call(x),
        mocker.call(x)
    ])

    _write_.reset_mock()

    # Test load_attr
    ctx = mocker.MagicMock()
    ctx.__enter__.return_value = ctx

    w = mocker.Mock(ctx=ctx)

    glb['load_attr'](w)

    assert w.ctx.z == 1
    _getattr_.assert_called_once_with(w, 'ctx')
    _write_.assert_called_once_with(w.ctx)


WITH_STMT_SUBSCRIPT = """
def single_key(ctx, x):
    with ctx as x['key']:
        pass


def slice_key(ctx, x):
    with ctx as x[2:3]:
        pass
"""


def test_with_stmt_subscript(mocker):
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob

    glb = {'_write_': _write_}
    restricted_exec(WITH_STMT_SUBSCRIPT, glb)

    # Test single_key
    ctx = mocker.MagicMock()
    ctx.__enter__.return_value = ctx
    x = {}

    glb['single_key'](ctx, x)

    assert x['key'] == ctx
    _write_.assert_called_once_with(x)
    _write_.reset_mock()

    # Test slice_key
    ctx = mocker.MagicMock()
    ctx.__enter__.return_value = (1, 2)

    x = [0, 0, 0, 0, 0, 0]
    glb['slice_key'](ctx, x)

    assert x == [0, 0, 1, 2, 0, 0, 0]
    _write_.assert_called_once_with(x)



from Aspidites._vendor._compat import basestring

from Aspidites._vendor.contracts.interface import (ContractSyntaxError, describe_value,
                                                   ContractNotRespected)
from Aspidites._vendor.contracts.main import parse_contract_string, check_contracts


def check_contracts_ok(contract, value):
    if isinstance(contract, basestring):
        contract = [contract]
        value = [value]
    context = check_contracts(contract, value)

    assert isinstance(context, dict)
    "%s" % context
    "%r" % context


def check_contracts_fail(contract, value, error=ContractNotRespected):
    """ Returns the exception """
    if isinstance(contract, basestring):
        contract = [contract]
        value = [value]

    try:
        context = check_contracts(contract, value)

        msg = ('I was expecting that the values would not not'
               ' satisfy the contract.\n')

        for v in value:
            msg += '      value: %s\n' % describe_value(v)

        for c in contract:
            cp = parse_contract_string(c)
            msg += '   contract: %r, parsed as %r (%s)\n' % (c, cp, cp)

        msg += '    context:  %r\n' % context

        raise Exception(msg)

    except error as e:
        # Try generation of strings:
        s = "%r" % e  # @UnusedVariable
        s = "%s" % e  # @UnusedVariable
        return e


def check_syntax_fail(string):
    assert isinstance(string, basestring)

    try:
        parsed_contract = parse_contract_string(string)
        msg = 'I would not expect to parse %r.' % string
        msg += ' contract:         %s\n' % parsed_contract
        raise Exception(msg)

    except ContractSyntaxError as e:
        # Try generation of strings:
        s = "%r" % e  # @UnusedVariable
        s = "%s" % e  # @UnusedVariable
        pass




import pytest as pt
from Aspidites.final import final, heritable
from Aspidites._vendor.contracts import ContractNotRespected


class A(metaclass=final):
    pass


class B:
    pass


def test_final_classes():
    with pt.raises(ContractNotRespected):
        class C(A):
            pass


def test_heritable_clause():
    assert not heritable((A,))
    assert heritable((B,))



from Aspidites._vendor.contracts import ContractNotRespected
from hypothesis import given, assume, strategies as st
import pytest as pt
from Aspidites.__main__ import get_cy_kwargs
from Aspidites.parser import parse_module
from Aspidites.templates import lib, setup
from Aspidites.monads import Maybe, Undefined, Surely
from Aspidites.compiler import compile_module

with open('Aspidites/tests/examples/examples.wom', 'r') as f:
    code_ = parse_module(f.read())


def test_compile_module():
    compile(lib.substitute(code='\n'.join(code_)), '', 'exec')


@given(x=st.integers())
def test_integer_monad(x):
    assert Undefined() == Undefined()
    assert Undefined() + Undefined() == Undefined()
    assert Undefined() + x == x
    assert Undefined() - Undefined() == Undefined()
    assert Undefined() - x == -x
    assert Undefined() * x == Undefined()
    assert Undefined() * Undefined() == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() / Undefined() == Undefined()
    assert Surely() == Surely()
    assert Surely() + Surely() == Surely()
    assert Surely() + x == x
    assert Surely() - Surely() == Surely()
    assert Surely() - x == -x
    assert Surely() * x == Surely()
    assert Surely() * Surely() == Surely()
    assert Surely() / x == Surely()
    assert Surely() / Surely() == Surely()
    assert Undefined(x) != Surely(x)
    assert Maybe(x) != x
    assert Surely(x) == Surely(x)
    assert Maybe(x) != Surely(x)
    assert Surely(Maybe(x)) != x
    assert Surely(Undefined()) == Undefined()
    assert Surely(x) == x
    assert -Surely(x) == -x
    assert ~Surely(x) == ~x
    assert ~Maybe(x) != ~x
    assert -Maybe(x) != -x
    assert ~Surely() == Undefined()
    assert oct(Surely(x)) == oct(x)


def test_compile_to_shared_object():

    compile_module(code_, 'Aspidites/tests/examples/compiled.py', bytecode=True, **get_cy_kwargs())

    from Aspidites.tests.examples.compiled import Add, x, y, z, scala, val, div_by_zero, Yield123, Hello, Hello2

    with pt.raises(ContractNotRespected):
        Add(x=6.5, y=12)

    assert [1, 2, 3] == [i for i in Yield123()]
    assert Maybe(Add, 6.5, 12)() == Undefined()
    assert x() == 6
    assert y() == Undefined()
    assert z == 9
    assert Add(x=3, y=2) == 5
    assert val() == Undefined()
    assert div_by_zero() == Undefined()
    Hello()


#
# def teardown_function():
#     for file in glob.glob('examples/compiled.*'):
#         os.remove(file)
#     try:
#         os.remove('examples/setup.py')
#     except FileNotFoundError:
#         pass



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8
# THIS FILE IS GENERATED - DO NOT EDIT #
import cython  # type: ignore
from typing import Any
from collections.abc import Generator
from pyrsistent import (
    pset, 
    pmap, 
    pvector, 
    s, v, m, 
    PRecord,
    PClass
)
from Aspidites.woma import *
from Aspidites._vendor import F, _
from Aspidites.monads import Maybe, Surely, Undefined, SafeDiv, SafeMod, SafeExp
from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.RestrictedPython import safe_builtins
# DECLARATIONS TO ALLOW CONTRACTS TO TYPE CHECK #
procedure: None
coroutine: Generator
number: Any
globals().update(dict(__builtins__=safe_builtins))  # add all imports to globals


# comment_line 1:`first-class functions`
@contract()
@cython.binding(True)
def Add(x : 'int' = 3, y : 'int' = 3) -> 'int':
    return x+y


# comment_line 6:`generators`
@contract()
def Yield123() -> 'coroutine':
    yield Maybe(Add, 0, 1)()
    yield Maybe(Add, 0, 2)()
    yield Maybe(Add, 0, 3)()


# comment_line 12:`procedures`
@contract()
def Hello() -> 'procedure':
    return Maybe(print, "Hello, World!")()


# comment_line 16:`coroutines`
@contract()
def Hello2() -> 'coroutine':
    yield Maybe(Hello, )()


# comment_line 20:`persistent vectors`
D = pvector([2, 4, 6, 8, 10])
# comment_line 23:`persistent sets`
G = pset({'a', 'b', 'c'})
# comment_line 26:`persistent mappings`
C = pmap({'a': (3+5), 'b': 8, 'c': True, 4: None, 'd': Maybe(SafeExp, 6, 2*5+3)})
# comment_line 33:`new contracts can impose more complex contractual clauses`
new_contract('colors', 'list[3](int,<256)')
# comment_line 36:`any woma function can be closed in place to become an instance that complies with the`
# comment_line 37:`type specification or Undefined for instances that breach the type specification contract`
x = Maybe(Add, 3, 3)
# comment_line 40:`seamless exception handling allows tracing of undefined code branches`
y = Maybe(Add, 4, 3.5)
# comment_line 43:`mixed usage of closure and regular function calls`
z = Maybe(Add, Maybe(x, )(), 3)()
# comment_line 46:`Scala-style closure functions`
scala = (_*2)
val = Maybe(scala, _+_)()
val = Maybe(val, scala)
# comment_line 51:`modulus and division by 0 handled by returning Undefined()`
denom = 0
div_by_zero = Maybe(SafeDiv, 1, denom)
mod_zero = Maybe(SafeMod, 1, denom)
div_by_zero2 = Maybe(SafeDiv, 1, 0)
mod_zero2 = Maybe(SafeMod, 1, 0)
# comment_line 58:`main: structure for executable actions when run as a binary`
if __name__ == "__main__":
    Maybe(Hello, )()
    Maybe(print, "I'm a binary.")()




from ..interface import Contract, ContractNotRespected
from ..syntax import (W, contract_expression, O, S, add_contract, add_keyword,
    Keyword)


class Dict(Contract):

    def __init__(self, length=None, key_c=None, value_c=None, where=None):
        Contract.__init__(self, where)
        self.length = length
        self.key_c = key_c
        self.value_c = value_c

    def check_contract(self, context, value, silent):
        if not isinstance(value, dict):
            error = 'Expected a dict, got %r.' % value.__class__.__name__
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        if self.length is not None:
            self.length._check_contract(context, len(value), silent)

        for k in value:
            if self.key_c is not None:
                self.key_c._check_contract(context, k, silent)
            if self.value_c is not None:
                self.value_c._check_contract(context, value[k], silent)

    def __str__(self):
        s = 'dict'
        if self.length is not None:
            s += '[%s]' % self.length
        if self.key_c is not None:
            k = str(self.key_c)
        else:
            k = ''
        if self.value_c is not None:
            v = str(self.value_c)
        else:
            v = ''
        if k or v:
            s += '(%s:%s)' % (k, v)
        return s

    def __repr__(self):
        return 'Dict(%r,%r,%r)' % (self.length, self.key_c, self.value_c)

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length = tokens.get('length', None)
        key = tokens.get('key', None)
        value = tokens.get('value', None)
        return Dict(length, key, value, where=where)


length_spec = S('[') - contract_expression('length') - S(']')
kv_spec = ('(' - O(contract_expression('key'))
           + ':' + O(contract_expression('value')) - ')')
dict_contract = Keyword('dict') + O(length_spec) + O(kv_spec)

dict_contract.setParseAction(Dict.parse_action)

add_contract(dict_contract)
add_keyword('dict')



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor._compat import IS_PY3

import pytest


EXEC_STATEMENT = """\
def no_exec():
    exec 'q = 1'
"""


@pytest.mark.skipif(IS_PY3,
                    reason="exec statement no longer exists in Python 3")
def test_RestrictingNodeTransformer__visit_Exec__1():  # pragma: PY2
    """It prevents using the `exec` statement. (Python 2 only)"""
    result = compile_restricted_exec(EXEC_STATEMENT)
    assert result.errors == ('Line 2: Exec statements are not allowed.',)


EXEC_FUNCTION = """\
def no_exec():
    exec('q = 1')
"""


@pytest.mark.skipif(IS_PY2,
                    reason="exec is a statement in Python 2")
def test_RestrictingNodeTransformer__visit_Exec__2():  # pragma: PY3
    """It is an error if the code call the `exec` function."""
    result = compile_restricted_exec(EXEC_FUNCTION)
    assert result.errors == ("Line 2: Exec calls are not allowed.",)


EVAL_FUNCTION = """\
def no_eval():
    eval('q = 1')
"""


def test_RestrictingNodeTransformer__visit_Eval__1():
    """It is an error if the code call the `eval` function."""
    result = compile_restricted_exec(EVAL_FUNCTION)
    assert result.errors == ("Line 2: Eval calls are not allowed.",)



import unittest

from Aspidites._vendor import match, ANY


class PampyDataClassesTests(unittest.TestCase):
    def test_dataclasses(self):
        try:
            from dataclasses import dataclass
        except ImportError:
            return

        @dataclass
        class Point:
            x: float
            y: float

        def f(x):
            return match(x,
                         Point(1, 2), '1',
                         Point(ANY, 2), str,
                         Point(1, ANY), str,
                         Point(ANY, ANY), lambda a, b: str(a + b)
                         )

        self.assertEqual(f(Point(1, 2)), '1')
        self.assertEqual(f(Point(2, 2)), '2')
        self.assertEqual(f(Point(1, 3)), '3')
        self.assertEqual(f(Point(2, 3)), '5')

    def test_different_dataclasses(self):
        try:
            from dataclasses import dataclass
        except ImportError:
            return

        @dataclass
        class Cat:
            name: str
            chassed_squirels: int

        @dataclass
        class Dog:
            name: str
            chassed_cats: int

        def what_is(x):
            return match(x,
                         Dog(ANY, 0), 'good boy',
                         Dog(ANY, ANY), 'doggy!',
                         Cat(ANY, 0), 'tommy?',
                         Cat(ANY, ANY), 'a cat'
                         )

        self.assertEqual(what_is(Cat("cat", 0)), 'tommy?')
        self.assertEqual(what_is(Cat("cat", 1)), 'a cat')
        self.assertEqual(what_is(Dog("", 0)), 'good boy')
        self.assertEqual(what_is(Dog("", 1)), 'doggy!')



import unittest
from collections.abc import Iterable

from Aspidites._vendor import HEAD, TAIL, MatchError, match_iterable as mi
from Aspidites._vendor import ANY


class IterableTests(unittest.TestCase):

    def setUp(self):
        # So we can wrap it later
        self.mi = mi

    def test_match_iterable(self):
        self.assertEqual(self.mi([1, 2, 3], [1, 2, 3]), (True, []))
        self.assertEqual(self.mi([str, str], ["hello", "world"]), (True, ["hello", "world"]))
        self.assertEqual(self.mi([1, 2], [1, 3]), (False, []))

        self.assertEqual(self.mi([1, 2], [1]), (False, []))

    def test_match_iterable_with_a_value(self):
        self.assertEqual(self.mi([1, 2], 33), (False, []))
        self.assertEqual(self.mi(33, [1, 2]), (False, []))

    def test_match_iterable_underscore(self):
        self.assertEqual(self.mi([1, ANY, 3], [1, 2, 3]), (True, [2]))
        self.assertEqual(self.mi([1, ANY, ANY], [1, 2, 3]), (True, [2, 3]))
        self.assertEqual(self.mi([ANY, 2, ANY], [1, 2, 3]), (True, [1, 3]))

    def test_match_iterable_pattern_longer_than_var(self):
        self.assertEqual(self.mi([ANY, ANY, ANY], [1, 2]), (False, []))
        self.assertEqual(self.mi([ANY, 2, ANY], [1, 2]), (False, []))

    def test_match_iterable_nested(self):
        self.assertEqual(self.mi([1, [2, 3]], [1, [2, 3]]), (True, []))
        self.assertEqual(self.mi([1, [2, ANY]], [1, [2, 3]]), (True, [3]))
        self.assertEqual(self.mi([1, [2, ANY], ANY],
                                 [1, [2, 3], 4]), (True, [3, 4]))
        self.assertEqual(self.mi([1, [2, [ANY, 4], ANY], ANY],
                                 [1, [2, [3, 4], 5], 6]), (True, [3, 5, 6]))

    def test_match_iterable_basic_HEAD(self):
        self.assertEqual(self.mi([HEAD], [1]),   (True, [1]))
        self.assertEqual(self.mi([HEAD], []),    (False, []))

    def test_match_iterable_HEAD(self):
        self.assertEqual(self.mi([HEAD, 2], [1, 2]), (True, [1]))
        self.assertEqual(self.mi([HEAD, ANY], [1, 2]), (True, [1, 2]))

    def test_match_iterable_TAIL(self):
        self.assertEqual(self.mi([TAIL],         [1, 2, 3]), (True, [[1, 2, 3]]))
        self.assertEqual(self.mi([ANY, TAIL],      [1, 2, 3]), (True, [1, [2, 3]]))
        self.assertEqual(self.mi([ANY, ANY, TAIL],   [1, 2, 3]), (True, [1, 2, [3]]))

        self.assertEqual(self.mi([ANY, TAIL],      [1]),       (True, [1, []]))
        self.assertEqual(self.mi([ANY, TAIL],      [1, 2]),    (True, [1, [2]]))
        self.assertEqual(self.mi([ANY, ANY, TAIL],   [1, 2]),    (True, [1, 2, []]))

    def test_match_iterable_HEAD_TAIL(self):
        self.assertEqual(self.mi([HEAD, TAIL], [1, 2, 3]),   (True, [1, [2, 3]]))
        self.assertEqual(self.mi([HEAD, TAIL], [1]),         (True, [1, []]))
        self.assertEqual(self.mi([HEAD, TAIL], []),          (False, []))

    def test_match_iterable_HEAD_should_be_in_first_position(self):
        with self.assertRaises(MatchError):
            self.mi([1, HEAD], [1, 2])

        with self.assertRaises(MatchError):
            self.mi([1, HEAD, 3], [1, 2, 3])

    def test_match_iterable_TAIL_should_be_in_last_position(self):
        with self.assertRaises(MatchError):
            self.mi([1, TAIL, 3], [1, 2, 3])

        with self.assertRaises(MatchError):
            self.mi([TAIL, 2], [1, 2, 3])

    def test_match_iterable_with_tuples(self):
        def convert_to_tuples(patterns, values):
            patterns = tuple(patterns) if isinstance(patterns, Iterable) else patterns
            values = tuple(values) if isinstance(values, Iterable) else values
            return mi(patterns, values)

        self.mi = convert_to_tuples

        methods = [f for f in dir(self) if callable(getattr(self, f))
                   if f.startswith('test_') and f != 'test_match_iterable_with_tuples']

        for method in methods:
            getattr(self, method)()

        self.mi = mi



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor.RestrictedPython.Guards import safe_builtins
from Aspidites.tests.helper import restricted_exec

import pytest


GOOD_CLASS = '''
class Good:
    pass
'''


def test_RestrictingNodeTransformer__visit_ClassDef__1():
    """It allows to define an class."""
    result = compile_restricted_exec(GOOD_CLASS)
    assert result.errors == ()
    assert result.code is not None


BAD_CLASS = '''\
class _bad:
    pass
'''


def test_RestrictingNodeTransformer__visit_ClassDef__2():
    """It does not allow class names which start with an underscore."""
    result = compile_restricted_exec(BAD_CLASS)
    assert result.errors == (
        'Line 1: "_bad" is an invalid variable name '
        'because it starts with "_"',)


IMPLICIT_METACLASS = '''
class Meta:
    pass

b = Meta().foo
'''


def test_RestrictingNodeTransformer__visit_ClassDef__3():
    """It applies the global __metaclass__ to all generated classes if present.
    """
    def _metaclass(name, bases, dict):
        ob = type(name, bases, dict)
        ob.foo = 2411
        return ob

    restricted_globals = dict(
        __metaclass__=_metaclass,
        __name__='implicit_metaclass',
        b=None,
        _getattr_=getattr)

    restricted_exec(IMPLICIT_METACLASS, restricted_globals)

    assert restricted_globals['b'] == 2411


EXPLICIT_METACLASS = '''
class WithMeta(metaclass=MyMetaClass):
    pass
'''


@pytest.mark.skipif(IS_PY2, reason="No valid syntax in Python 2.")
def test_RestrictingNodeTransformer__visit_ClassDef__4():
    """It does not allow to pass a metaclass to class definitions."""

    result = compile_restricted_exec(EXPLICIT_METACLASS)

    assert result.errors == (
        'Line 2: The keyword argument "metaclass" is not allowed.',)
    assert result.code is None


DECORATED_CLASS = '''\
def wrap(cls):
    cls.wrap_att = 23
    return cls

class Base:
    base_att = 42

@wrap
class Combined(Base):
    class_att = 2342

comb = Combined()
'''


def test_RestrictingNodeTransformer__visit_ClassDef__5():
    """It preserves base classes and decorators for classes."""

    restricted_globals = dict(
        comb=None, _getattr_=getattr, _write_=lambda x: x, __metaclass__=type,
        __name__='restricted_module', __builtins__=safe_builtins)

    restricted_exec(DECORATED_CLASS, restricted_globals)

    comb = restricted_globals['comb']
    assert comb.class_att == 2342
    assert comb.base_att == 42
    assert comb.wrap_att == 23


CONSTRUCTOR_TEST = """\
class Test:
    def __init__(self, input):
        self.input = input

t = Test(42)
"""


def test_RestrictingNodeTransformer__visit_ClassDef__6():
    """It allows to define an ``__init__`` method."""
    restricted_globals = dict(
        t=None,
        _write_=lambda x: x,
        __metaclass__=type,
        __name__='constructor_test',
    )

    restricted_exec(CONSTRUCTOR_TEST, restricted_globals)
    t = restricted_globals['t']
    assert t.input == 42


COMPARE_TEST = """\
class Test:

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

a = Test(42)
b = Test(42)
c = Test(43)

result1 = (a == b)
result2 = (b == c)
"""


def test_RestrictingNodeTransformer__visit_ClassDef__7():
    """It allows to define an ``__eq__`` method."""
    restricted_globals = dict(
        result1=None,
        result2=None,
        _getattr_=getattr,
        _write_=lambda x: x,
        __metaclass__=type,
        __name__='compare_test',
    )

    restricted_exec(COMPARE_TEST, restricted_globals)
    assert restricted_globals['result1'] is True
    assert restricted_globals['result2'] is False


CONTAINER_TEST = """\
class Test(object):

    def __init__(self, values):
        self.values = values

    def __contains__(self, value):
        return value in self.values

a = Test([1, 2, 3])

result1 = (1 in a)
result2 = (4 not in a)
"""


def test_RestrictingNodeTransformer__visit_ClassDef__8():
    """It allows to define a ``__contains__`` method."""
    restricted_globals = dict(
        result1=None,
        result2=None,
        _getattr_=getattr,
        _write_=lambda x: x,
        __metaclass__=type,
        __name__='container_test',
        object=object,
    )

    restricted_exec(CONTAINER_TEST, restricted_globals)
    assert restricted_globals['result1'] is True
    assert restricted_globals['result2'] is True



import sys
import inspect
import threading
import time
import functools
import itertools
import collections
import collections.abc as c
from decorator import (decorator, decorate, FunctionMaker,
                       dispatch_on, __version__)

doc = r"""Decorators for Humans
----------------------------------

|Author | Michele Simionato|
|---|---|
|E-mail | michele.simionato@gmail.com|
|Version| $VERSION ($DATE)|
|Supports| Python 3.5, 3.6, 3.7, 3.8, 3.9|
|Download page| http://pypi.python.org/pypi/decorator/$VERSION|
|Installation| ``pip install decorator``|
|License | BSD license|

Introduction
-----------------------------------------

The ``decorator`` module is over ten years old, but still alive and
kicking. It is used by several frameworks (IPython, scipy, authkit,
pylons, pycuda, sugar, ...) and has been stable for a *long* time. It
is your best option if you want to preserve the signature of decorated
functions in a consistent way across Python releases. Versions 5.X
supports Python versions greater than 3.4, versions 4.X supports Python
versions back to 2.6; versions 3.X are able to support even Python 2.5 and
2.4.

What's New in version 5
-----------------------

Version 5 of the decorator module features a major simplification of
the code base made possible by dropping support for Python releases
older than 3.5. From that version the ``Signature`` object works well
enough that it is possible to fix the signature of a decorated
function without resorting to ``exec`` tricks. The simplification
has a very neat advantage: in case of exceptions raised in decorated
functions the traceback is nicer than it used to be. Moreover, it is
now possible to mimic the behavior of decorators defined with
``functool.wraps``: see the section about the ``kwsyntax`` flag below.

What's New in version 4
-----------------------

- **New documentation**
  There is now a single manual for all Python versions, so I took the
  opportunity to overhaul the documentation.
  Even if you are a long-time user, you may want to revisit the docs, since
  several examples have been improved.

- **Packaging improvements**
  The code is now also available in wheel format. Integration with
  setuptools has improved and you can run the tests with the command
  ``python setup.py test`` too.

- **Code changes**
  A new utility function ``decorate(func, caller)`` has been added.
  It does the same job that was performed by the older
  ``decorator(caller, func)``. The old functionality is now deprecated
  and no longer documented, but still available for now.

- **Multiple dispatch**
  The decorator module now includes an implementation of generic
  functions (sometimes called "multiple dispatch functions").
  The API is designed to mimic ``functools.singledispatch`` (added
  in Python 3.4), but the implementation is much simpler.
  Moreover, all decorators involved preserve the signature of the
  decorated functions. For now, this exists mostly to demonstrate
  the power of the module. In the future it could be enhanced/optimized.
  In any case, it is very short and compact (less then 100 lines), so you
  can extract it for your own use. Take it as food for thought.

- **Python 3.5 coroutines**
  From version 4.1 it is possible to decorate coroutines, i.e. functions
  defined with the `async def` syntax, and to maintain the
  `inspect.iscoroutinefunction` check working for the decorated function.

- **Decorator factories**
  From version 4.2 there is facility to define factories of decorators in
  a simple way, a feature requested by the users since a long time.

Usefulness of decorators
------------------------------------------------

Python decorators are an interesting example of why syntactic sugar
matters. In principle, their introduction in Python 2.4 changed
nothing, since they did not provide any new functionality which was not
already present in the language. In practice, their introduction has
significantly changed the way we structure our programs.
I believe the change is for the best, and that decorators are a great
idea since:

* decorators help reducing boilerplate code;
* decorators help separation of concerns;
* decorators enhance readability and maintenability;
* decorators are explicit.

Still, as of now, writing custom decorators correctly requires
some experience and it is not as easy as it could be. For instance,
typical implementations of decorators involve nested functions, and
we all know that flat is better than nested.

The aim of the ``decorator`` module it to simplify the usage of
decorators for the average programmer, and to popularize decorators by
showing various non-trivial examples. Of course, as all techniques,
decorators can be abused (I have seen that) and you should not try to
solve every problem with a decorator, just because you can.

You may find the source code for all the examples
discussed here in the ``documentation.py`` file, which contains
the documentation you are reading in the form of doctests.

Definitions
------------------------------------

Technically speaking, any Python object which can be called with one argument
can be used as a decorator. However, this definition is somewhat too large
to be really useful. It is more convenient to split the generic class of
decorators in two subclasses:

1. **signature-preserving decorators**, callable objects which accept
    a function as input and return a function as output, *with the
    same signature*

2. **signature-changing** decorators, i.e. decorators
    which change the signature of their input function, or decorators
    that return non-callable objects

Signature-changing decorators have their use: for instance, the
builtin classes ``staticmethod`` and ``classmethod`` are in this
group. They take functions and return descriptor objects which
are neither functions, nor callables.

Still, signature-preserving decorators are more common, and easier
to reason about. In particular, they can be composed together,
whereas other decorators generally cannot.

Writing signature-preserving decorators from scratch is not that
obvious, especially if one wants to define proper decorators that
can accept functions with any signature. A simple example will clarify
the issue.

Statement of the problem
------------------------------

A very common use case for decorators is the memoization of functions.
A ``memoize`` decorator works by caching
the result of the function call in a dictionary, so that the next time
the function is called with the same input parameters the result is retrieved
from the cache and not recomputed.

There are many implementations of ``memoize`` in
http://www.python.org/moin/PythonDecoratorLibrary,
but they do not preserve the signature. In recent versions of
Python you can find a sophisticated ``lru_cache`` decorator
in the standard library's ``functools``. Here I am just
interested in giving an example.

Consider the following simple implementation (note that it is
generally impossible to *correctly* memoize something
that depends on non-hashable arguments):

$$memoize_uw

Here I used the functools.update_wrapper_ utility, which was added
in Python 2.5 to simplify the writing of decorators.
(Previously, you needed to manually copy the function attributes
``__name__``, ``__doc__``, ``__module__``, and ``__dict__``
to the decorated function by hand).

This works insofar as the decorator accepts functions with generic signatures.
Unfortunately, it is *not* a signature-preserving decorator, since
``memoize_uw`` generally returns a function with a *different signature*
from the original.

Consider for instance the following case:

$$f1

Here, the original function takes a single argument named ``x``,
but the decorated function takes any number of arguments and
keyword arguments:

```python
>>> from inspect import getfullargspec
>>> print(getfullargspec(f1))
FullArgSpec(args=[], varargs='args', varkw='kw', defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})

```

This means that introspection tools like ``getfullargspec`` will give
you false information about the signature of ``f1`` This is pretty bad:
``getfullargspec`` says that the function accepts the generic
signature ``*args, **kw``, but calling the function with more than one
argument raises an error:

```python
>>> f1(0, 1) # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
   ...
TypeError: f1() takes exactly 1 positional argument (2 given)

```

Notice that ``pydoc`` will give the right signature, but only in Python
versions greater than 3.5.

The solution
-----------------------------------------

The solution is to provide a generic factory of generators, which
hides the complexity of making signature-preserving decorators
from the application programmer. The ``decorate`` function in
the ``decorator`` module is such a factory:

```python
>>> from decorator import decorate

```

``decorate`` takes two arguments:

1. a caller function describing the functionality of the decorator, and

2. a function to be decorated.

The caller function must have signature ``(f, *args, **kw)``, and it
must call the original function ``f`` with arguments ``args`` and ``kw``,
implementing the wanted capability (in this case, memoization):

$$_memoize

Now, you can define your decorator as follows:

$$memoize

The difference from the nested function approach of ``memoize_uw``
is that the decorator module forces you to lift the inner function
to the outer level. Moreover, you are forced to explicitly pass the
function you want to decorate; there are no closures.

Here is a test of usage:

```python
>>> @memoize
... def heavy_computation():
...     time.sleep(2)
...     return "done"

>>> print(heavy_computation()) # the first time it will take 2 seconds
done

>>> print(heavy_computation()) # the second time it will be instantaneous
done

```

The signature of ``heavy_computation`` is the one you would expect:

```python
>>> print(getfullargspec(heavy_computation))
FullArgSpec(args=[], varargs=None, varkw=None, defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})

```

A ``trace`` decorator
------------------------------------------------------

Here is an example of how to define a simple ``trace`` decorator,
which prints a message whenever the traced function is called:

$$_trace

$$trace

Here is an example of usage:

```python
>>> @trace
... def f1(x):
...     pass

```

It is immediate to verify that ``f1`` works...

```python
>>> f1(0)
calling f1 with args (0,), {}

```

...and it that it has the correct signature:

```python
>>> print(getfullargspec(f1))
FullArgSpec(args=['x'], varargs=None, varkw=None, defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})

```

The decorator works with functions of any signature:

```python
>>> @trace
... def f(x, y=1, *args, **kw):
...     pass

>>> f(0, 3)
calling f with args (0, 3), {}

>>> print(getfullargspec(f))
FullArgSpec(args=['x', 'y'], varargs='args', varkw='kw', defaults=(1,), kwonlyargs=[], kwonlydefaults=None, annotations={})

```

$FUNCTION_ANNOTATIONS

``decorator.decorator``
---------------------------------------------

It can become tedious to write a caller function (like the above
``_trace`` example) and then a trivial wrapper
(``def trace(f): return decorate(f, _trace)``) every time.
Not to worry!  The ``decorator`` module provides an easy shortcut
to convert the caller function into a signature-preserving decorator.

It is the ``decorator`` function:

```python
>>> from decorator import decorator

```
The ``decorator`` function can be used as a signature-changing
decorator, just like ``classmethod`` and ``staticmethod``.
But ``classmethod`` and ``staticmethod`` return generic
objects which are not callable. Instead, ``decorator`` returns
signature-preserving decorators (i.e. functions with a single argument).

For instance, you can write:

```python
>>> @decorator
... def trace(f, *args, **kw):
...     kwstr = ', '.join('%r: %r' % (k, kw[k]) for k in sorted(kw))
...     print("calling %s with args %s, {%s}" % (f.__name__, args, kwstr))
...     return f(*args, **kw)

```

And ``trace`` is now a decorator!

```python
>>> trace # doctest: +ELLIPSIS
<function trace at 0x...>

```

Here is an example of usage:

```python
>>> @trace
... def func(): pass

>>> func()
calling func with args (), {}

```

Mimicking the behavior of functools.wrap
----------------------------------------

Often people are confused by the decorator module since, contrarily
to ``functools.wraps`` in the standard library, it tries very hard
to keep the semantics of the arguments: in particular, positional arguments
stay positional even if they are called with the keyword argument syntax.
An example will make the issue clear. Here is a simple caller

$$chatty

and here is a function to decorate:

$$printsum

In this example ``x`` and ``y`` are positional arguments (with
defaults). From the caller perspective, it does not matter if the user
calls them as named arguments, they will stay inside the ``args``
tuple and not inside the ``kwargs`` dictionary:

```python
>>> printsum(y=2, x=1)
(1, 2) []
3

```

This is quite different from the behavior of ``functools.wraps``; if you
define the decorator as follows

$$chattywrapper

you will see that calling ``printsum`` with named arguments will pass
such arguments to ``kwargs``, while ``args`` will be the empty tuple.
Since version 5 of the decorator module it is possible to mimic that
behavior by using the ``kwsyntax`` flag:

$$printsum2

Here is how it works:

```python
>>> printsum2(y=2, x=1)
() [('x', 1), ('y', 2)]
3

```

This is exactly what the ``chattywrapper`` decorator would print:
positional arguments are seen as keyword arguments, but only if the
client code calls them with the keyword syntax. Otherwise they stay
positional, i.e. they belongs to the ``args`` tuple and not to ``kwargs``:

```python
>>> printsum2(1, 2)
(1, 2) []
3

```

Decorator factories
-------------------------------------------

The `decorator` function can also be used to define factories of decorators,
i.e. functions returning decorators. In general you can just write something
like this:

```python
def decfactory(param1, param2, ...):
    def caller(f, *args, **kw):
        return somefunc(f, param1, param2, .., *args, **kw)
    return decorator(caller)
```

This is fully general but requires an additional level of nesting. For this
reason since version 4.2 there is a facility to build decorator factories by
using a single caller with default arguments:

```python
def caller(f, param1=default1, param2=default2, ..., *args, **kw):
    return somefunc(f, param1, param2, *args, **kw)
decfactory = decorator(caller)
```

Notice that this simplified approach *only works with default arguments*,
i.e. `param1`, `param2` etc must have known defaults. Thanks to this
restriction, there exists an unique default decorator, i.e. the member
of the family which uses the default values for all parameters. Such
decorator can be written as ``decfactory()`` with no parameters specified;
moreover, as a shortcut, it is also possible to elide the parenthesis,
a feature much requested by the users. For years I have been opposing
the request, since having explicit parenthesis to me is more clear
and less magic; however once this feature entered in decorators of
the Python standard library (I am referring to the [dataclass decorator](
https://www.python.org/dev/peps/pep-0557/)) I finally gave up.

The example below shows how it works in practice. The goal is to
convert a function relying on a blocking resource into a function
returning a "busy" message if the resource is not available.
This can be accomplished with a suitable family of decorators
parameterize by a string, the busy message:

$$blocking

Functions decorated with ``blocking`` will return a busy message if
the resource is unavailable, and the intended result if the resource is
available. For instance:

```python
>>> @blocking(msg="Please wait ...")
... def read_data():
...     time.sleep(3) # simulate a blocking resource
...     return "some data"

>>> print(read_data())  # data is not available yet
Please wait ...

>>> time.sleep(1)
>>> print(read_data())  # data is not available yet
Please wait ...

>>> time.sleep(1)
>>> print(read_data())  # data is not available yet
Please wait ...

>>> time.sleep(1.1)  # after 3.1 seconds, data is available
>>> print(read_data())
some data

```

Decorator factories are most useful to framework builders. Here is an example
that gives an idea of how you could manage permissions in a framework:

$$Action

where ``restricted`` is a decorator factory defined as follows

$$restricted

Notice that if you forget to use the keyword argument notation, i.e. if you
write ``restricted(User)`` instead of ``restricted(user_class=User)`` you
will get an error

```python
TypeError: You are decorating a non function: <class '__main__.User'>

```

Be careful!

``decorator(cls)``
--------------------------------------------

The ``decorator`` facility can also produce a decorator starting
from a class with the signature of a caller. In such a case the
produced generator is able to convert functions into factories
to create instances of that class.

As an example, here is a decorator which can convert a
blocking function into an asynchronous function. When
the function is called, it is executed in a separate thread.

(This is similar to the approach used in the ``concurrent.futures`` package.
But I don't recommend that you implement futures this way; this is just an
example.)

$$Future

The decorated function returns a ``Future`` object. It has a ``.result()``
method which blocks until the underlying thread finishes and returns
the final result.

Here is the minimalistic usage:

```python
>>> @decorator(Future)
... def long_running(x):
...     time.sleep(.5)
...     return x

>>> fut1 = long_running(1)
>>> fut2 = long_running(2)
>>> fut1.result() + fut2.result()
3

```

contextmanager
-------------------------------------

Python's standard library has the ``contextmanager`` decorator,
which converts a generator function into a ``GeneratorContextManager``
factory. For instance, if you write this...

```python
>>> from contextlib import contextmanager
>>> @contextmanager
... def before_after(before, after):
...     print(before)
...     yield
...     print(after)

```

...then ``before_after`` is a factory function that returns
``GeneratorContextManager`` objects, usable with the ``with`` statement:

```python
>>> with before_after('BEFORE', 'AFTER'):
...     print('hello')
BEFORE
hello
AFTER

```

Basically, it is as if the content of the ``with`` block was executed
in the place of the ``yield`` expression in the generator function.

In Python 3.2, ``GeneratorContextManager`` objects were enhanced with
a ``__call__`` method, so that they can be used as decorators, like so:

```python
>>> ba = before_after('BEFORE', 'AFTER')
>>>
>>> @ba # doctest: +SKIP
... def hello():
...     print('hello')
...
>>> hello() # doctest: +SKIP
BEFORE
hello
AFTER

```

The ``ba`` decorator basically inserts a ``with ba:`` block
inside the function.

However ``GeneratorContextManager`` objects do not preserve the signature of
the decorated functions. The decorated ``hello`` function above will
have the generic signature ``hello(*args, **kwargs)``, but fails if
called with more than zero arguments.

For these reasons, the `decorator` module, starting from release 3.4, offers a
``decorator.contextmanager`` decorator that solves both problems,
*and* works in all supported Python versions.  Its usage is identical,
and factories decorated with ``decorator.contextmanager`` will return
instances of ``ContextManager``, a subclass of the standard library's
``contextlib.GeneratorContextManager`` class. The subclass includes
an improved ``__call__`` method, which acts as a signature-preserving
decorator.

The ``FunctionMaker`` class
---------------------------------------------------------------

The ``decorator`` module also provides a ``FunctionMaker`` class, which
is able to generate on-the-fly functions
with a given name and signature from a function template
passed as a string.

If you're just writing ordinary decorators, then you probably won't
need to use ``FunctionMaker``. But in some circumstances, it
can be handy. You will see an example shortly--in
the implementation of a cool decorator utility (``decorator_apply``).

``FunctionMaker`` provides the ``.create`` classmethod, which
accepts the *name*, *signature*, and *body* of the function
you want to generate, as well as the execution environment
where the function is generated by ``exec``.

Here's an example:

```python
>>> def f(*args, **kw): # a function with a generic signature
...     print(args, kw)

>>> f1 = FunctionMaker.create('f1(a, b)', 'f(a, b)', dict(f=f))
>>> f1(1,2)
(1, 2) {}

```

It is important to notice that the function body is interpolated
before being executed; **be careful** with the ``%`` sign!

``FunctionMaker.create`` also accepts keyword arguments.
The keyword arguments are attached to the generated function.
This is useful if you want to set some function attributes
(e.g., the docstring ``__doc__``).

For debugging/introspection purposes, it may be useful to see
the source code of the generated function. To do this, just
pass ``addsource=True``, and the generated function will get
a ``__source__`` attribute:

```python
>>> f1 = FunctionMaker.create(
...     'f1(a, b)', 'f(a, b)', dict(f=f), addsource=True)
>>> print(f1.__source__)
def f1(a, b):
    f(a, b)
<BLANKLINE>

```

The first argument to ``FunctionMaker.create`` can be a string (as above),
or a function. This is the most common usage, since you typically decorate
pre-existing functions.

If you're writing a framework, however, you may want to use
``FunctionMaker.create`` directly, rather than ``decorator``, because it gives
you direct access to the body of the generated function.

For instance, suppose you want to instrument the ``__init__`` methods of a
set of classes, by preserving their signature.
(This use case is not made up. This is done by SQAlchemy, and other frameworks,
too.)
Here is what happens:

- If first argument of ``FunctionMaker.create`` is a function,
  an instance of ``FunctionMaker`` is created with the attributes
  ``args``, ``varargs``, ``keywords``, and ``defaults``
  (these mirror the return values of the standard library's
  ``inspect.getfullargspec``).

- For each item in ``args`` (a list of strings of the names of all required
  arguments), an attribute ``arg0``, ``arg1``, ..., ``argN`` is also generated.

- Finally, there is a ``signature`` attribute, which is a string with the
  signature of the original function.

**NOTE:** You should not pass signature strings with default arguments
(e.g., something like ``'f1(a, b=None)'``). Just pass ``'f1(a, b)'``,
followed by a tuple of defaults:

```python
>>> f1 = FunctionMaker.create(
...     'f1(a, b)', 'f(a, b)', dict(f=f), addsource=True, defaults=(None,))
>>> print(getfullargspec(f1))
FullArgSpec(args=['a', 'b'], varargs=None, varkw=None, defaults=(None,), kwonlyargs=[], kwonlydefaults=None, annotations={})

```

Getting the source code
---------------------------------------------------

Internally, ``FunctionMaker.create`` uses ``exec`` to generate the
decorated function. Therefore ``inspect.getsource`` will not work for
decorated functions. In IPython, this means that the usual ``??`` trick
will give you the (right on the spot) message ``Dynamically generated
function. No source code available``.
However, there is a workaround. The decorated function has the ``__wrapped__``
attribute, pointing to the original function. The simplest way to get the
source code is to call ``inspect.getsource`` on the undecorated function:

```python
>>> print(inspect.getsource(factorial.__wrapped__))
@tail_recursive
def factorial(n, acc=1):
    "The good old factorial"
    if n == 0:
        return acc
    return factorial(n-1, n*acc)
<BLANKLINE>

```

Dealing with third-party decorators
-----------------------------------------------------------------

Sometimes on the net you find some cool decorator that you would
like to include in your code. However, more often than not, the cool
decorator is not signature-preserving. What you need is an easy way to
upgrade third party decorators to signature-preserving decorators...
*without* having to rewrite them in terms of ``decorator``.

You can use a ``FunctionMaker`` to implement that functionality as follows:

$$decorator_apply

``decorator_apply`` sets the generated function's ``__wrapped__`` attribute
to the original function, so you can get the right source code.
If you are using a Python later than 3.2, you should also set the
``__qualname__`` attribute to preserve the qualified name of the original
function.

Notice that I am not providing this functionality in the ``decorator``
module directly, since I think it is best to rewrite the decorator instead
of adding another level of indirection. However, practicality
beats purity, so you can add ``decorator_apply`` to your toolbox and
use it if you need to.

To give a good example for ``decorator_apply``, I will show a pretty slick
decorator that converts a tail-recursive function into an iterative function.
I have shamelessly stolen the core concept from Kay Schluehr's recipe
in the Python Cookbook,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496691.

$$TailRecursive

Here the decorator is implemented as a class returning callable
objects.

$$tail_recursive

Here is how you apply the upgraded decorator to the good old factorial:

$$factorial

```python
>>> print(factorial(4))
24

```

This decorator is pretty impressive, and should give you some food for
thought! ;)

Notice that there is no recursion limit now; you can easily compute
``factorial(1001)`` (or larger) without filling the stack frame.

Notice also that the decorator will *not* work on functions which
are not tail recursive, such as the following:

$$fact

**Reminder:** A function is *tail recursive* if it does either of the
following:

- returns a value without making a recursive call; or,
- returns directly the result of a recursive call.

Python 3.5 coroutines
-----------------------

I am personally not using Python 3.5 coroutines yet. However, some
users requested support for coroutines and since version 4.1 the
decorator module has it.  You should consider the support experimental
and kindly report issues if you find any.

Here I will give a single example of usage. Suppose you want to log the moment
a coroutine starts and the moment it stops for debugging purposes. You could
write code like the following:

```python
import time
import logging
from asyncio import get_event_loop, sleep, wait
from decorator import decorator

@decorator
async def log_start_stop(coro, *args, **kwargs):
    logging.info('Starting %s%s', coro.__name__, args)
    t0 = time.time()
    await coro(*args, **kwargs)
    dt = time.time() - t0
    logging.info('Ending %s%s after %d seconds', coro.__name__, args, dt)

@log_start_stop
async def make_task(n):
    for i in range(n):
        await sleep(1)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    tasks = [make_task(3), make_task(2), make_task(1)]
    get_event_loop().run_until_complete(wait(tasks))
```

and you will get an output like this:

```bash
INFO:root:Starting make_task(1,)
INFO:root:Starting make_task(3,)
INFO:root:Starting make_task(2,)
INFO:root:Ending make_task(1,) after 1 seconds
INFO:root:Ending make_task(2,) after 2 seconds
INFO:root:Ending make_task(3,) after 3 seconds
```

This may be handy if you have trouble understanding what it going on
with a particularly complex chain of coroutines. With a single line you
can decorate the troubling coroutine function, understand what happens, fix the
issue and then remove the decorator (or keep it if continuous monitoring
of the coroutines makes sense). Notice that
``inspect.iscoroutinefunction(make_task)``
will return the right answer (i.e. ``True``).

It is also possible to define decorators converting coroutine functions
into regular functions, such as the following:

```python
@decorator
def coro_to_func(coro, *args, **kw):
    "Convert a coroutine into a function"
     return get_event_loop().run_until_complete(coro(*args, **kw))
```

Notice the difference: the caller in ``log_start_stop`` was a coroutine
function and the associate decorator was converting coroutines in coroutines;
the caller in ``coro_to_func`` is a regular function and converts
coroutines -> functions.

Multiple dispatch
-------------------------------------------

There has been talk of implementing multiple dispatch functions
(i.e. "generic functions") in Python for over ten years. Last year,
something concrete was done for the first time. As of Python 3.4,
we have the decorator ``functools.singledispatch`` to implement generic
functions!

As its name implies, it is limited to *single dispatch*; in other words,
it is able to dispatch on the first argument of the function only.

The ``decorator`` module provides the decorator factory ``dispatch_on``,
which can be used to implement generic functions dispatching on *any* argument.
Moreover, it can manage dispatching on more than one argument.
(And, of course, it is signature-preserving.)

Here is a concrete example (from a real-life use case) where it is desiderable
to dispatch on the second argument.

Suppose you have an ``XMLWriter`` class, which is instantiated
with some configuration parameters, and has the ``.write`` method which
serializes objects to XML:

$$XMLWriter

Here, you want to dispatch on the *second* argument; the first is already
taken by ``self``. The ``dispatch_on`` decorator factory allows you to specify
the dispatch argument simply by passing its name as a string. (Note
that if you misspell the name you will get an error.)

The decorated function `write` is turned into a generic function (
`write` is a function at the idea it is decorated; it will be turned
into a method later, at class instantiation time),
and it is called if there are no more specialized implementations.

Usually, default functions should raise a ``NotImplementedError``, thus
forcing people to register some implementation.
You can perform the registration with a decorator:

$$writefloat

Now ``XMLWriter`` can serialize floats:

```python
>>> writer = XMLWriter()
>>> writer.write(2.3)
'<float>2.3</float>'

```

I could give a down-to-earth example of situations in which it is desiderable
to dispatch on more than one argument--for instance, I once implemented
a database-access library where the first dispatching argument was the
the database driver, and the second was the database record--but here
I will follow tradition, and show the time-honored Rock-Paper-Scissors example:

$$Rock
$$Paper
$$Scissors

I have added an ordinal to the Rock-Paper-Scissors classes to simplify
the implementation. The idea is to define a generic function (``win(a,
b)``) of two arguments corresponding to the *moves* of the first and
second players. The *moves* are instances of the classes
Rock, Paper, and Scissors:

- Paper wins over Rock
- Scissors wins over Paper
- Rock wins over Scissors

The function will return +1 for a win, -1 for a loss, and 0 for parity.
There are 9 combinations, but combinations with the same ordinal
(i.e. the same class) return 0. Moreover, by exchanging the order of the
arguments, the sign of the result changes. Therefore, it is sufficient to
directly specify only 3 implementations:

$$win
$$winRockPaper
$$winPaperScissors
$$winRockScissors

Here is the result:

```python
>>> win(Paper(), Rock())
1
>>> win(Scissors(), Paper())
1
>>> win(Rock(), Scissors())
1
>>> win(Paper(), Paper())
0
>>> win(Rock(), Rock())
0
>>> win(Scissors(), Scissors())
0
>>> win(Rock(), Paper())
-1
>>> win(Paper(), Scissors())
-1
>>> win(Scissors(), Rock())
-1

```

The point of generic functions is that they play well with subclassing.
For instance, suppose we define a ``StrongRock``, which does not lose against
Paper:

$$StrongRock
$$winStrongRockPaper

Then you do not need to define other implementations; they are
inherited from the parent:

```python
>>> win(StrongRock(), Scissors())
1

```

You can introspect the precedence used by the dispatch algorithm by
calling ``.dispatch_info(*types)``:

```python
>>> win.dispatch_info(StrongRock, Scissors)
[('StrongRock', 'Scissors'), ('Rock', 'Scissors')]

```

Since there is no direct implementation for (``StrongRock``, ``Scissors``),
the dispatcher will look at the implementation for (``Rock``, ``Scissors``)
which is available. Internally, the algorithm is doing a cross
product of the class precedence lists (or *Method Resolution Orders*,
[MRO](http://www.python.org/2.3/mro.html) for short) of ``StrongRock``
 and ``Scissors``, respectively.

Generic functions and virtual ancestors
-------------------------------------------------

In Python, generic functions are complicated by the existence of
"virtual ancestors": superclasses which are not in the class hierarchy.

Consider this class:

$$WithLength

This class defines a ``__len__`` method, and is therefore
considered to be a subclass of the abstract base class
``collections.abc.Sized`` (``collections.Sized`` on Python 2):

```python
>>> issubclass(WithLength, collections.abc.Sized)
True

```

However, ``collections.abc.Sized`` is not in the MRO_ of ``WithLength``; it
is not a true ancestor. Any implementation of generic functions (even
with single dispatch) must go through some contorsion to take into
account the virtual ancestors.

In particular, if we define a generic function...

$$get_length

...implemented on all classes with a length...

$$get_length_sized

...then ``get_length`` must be defined on ``WithLength`` instances...

```python
>>> get_length(WithLength())
0

```

...even if ``collections.abc.Sized`` is not a true ancestor of ``WithLength``.

Of course, this is a contrived example--you could just use the
builtin ``len``--but you should get the idea.

Since in Python it is possible to consider any instance of ``ABCMeta``
as a virtual ancestor of any other class (it is enough to register it
as ``ancestor.register(cls)``), any implementation of generic functions
must be aware of the registration mechanism.

For example, suppose you are using a third-party set-like class, like
the following:

$$SomeSet

Here, the author of ``SomeSet`` made a mistake by inheriting from
``collections.abc.Sized`` (instead of ``collections.abc.Set``).

This is not a problem. You can register *a posteriori*
``collections.abc.Set`` as a virtual ancestor of ``SomeSet``:

```python
>>> _ = collections.abc.Set.register(SomeSet)
>>> issubclass(SomeSet, collections.abc.Set)
True

```

Now, let's define an implementation of ``get_length`` specific to set:

$$get_length_set

The current implementation (and ``functools.singledispatch`` too)
is able to discern that a ``Set`` is a ``Sized`` object, by looking at
the class registry, so it uses the more specific implementation for ``Set``:

```python
>>> get_length(SomeSet())  # NB: the implementation for Sized would give 0
1

```

Sometimes it is not clear how to dispatch. For instance, consider a
class ``C`` registered both as ``collections.abc.Iterable`` and
``collections.abc.Sized``, and defines a generic function ``g`` with
implementations for both ``collections.abc.Iterable`` *and*
``collections.abc.Sized``:

$$singledispatch_example1

It is impossible to decide which implementation to use, since the ancestors
are independent. The following function will raise a ``RuntimeError``
when called. This is consistent with the "refuse the temptation to guess"
philosophy. ``functools.singledispatch`` would raise a similar error.

It would be easy to rely on the order of registration to decide the
precedence order. This is reasonable, but also fragile:

- if, during some refactoring, you change the registration order by mistake,
  a different implementation could be taken;
- if implementations of the generic functions are distributed across modules,
  and you change the import order, a different implementation could be taken.

So the ``decorator`` module prefers to raise an error in the face of ambiguity.
This is the same approach taken by the standard library.

However, it should be noted that the *dispatch algorithm* used by the decorator
module is different from the one used by the standard library, so in certain
cases you will get different answers. The difference is that
``functools.singledispatch`` tries to insert the virtual ancestors *before* the
base classes, whereas ``decorator.dispatch_on`` tries to insert them *after*
the base classes.

Here's an example that shows the difference:

$$singledispatch_example2

If you play with this example and replace the ``singledispatch`` definition
with ``functools.singledispatch``, the assertion will break: ``g`` will return
``"container"`` instead of ``"s"``, because ``functools.singledispatch``
will insert the ``Container`` class right before ``S``.

Notice that here I am not making any bold claim such as "the standard
library algorithm is wrong and my algorithm is right" or viceversa. It
just point out that there are some subtle differences. The only way to
understand what is really happening here is to scratch your head by
looking at the implementations. I will just notice that
``.dispatch_info`` is quite essential to see the class precedence
list used by algorithm:

```python
>>> g, V = singledispatch_example2()
>>> g.dispatch_info(V)
[('V',), ('Sized',), ('S',), ('Container',)]

```

The current implementation does not implement any kind of cooperation
between implementations. In other words, nothing is akin either to
call-next-method in Lisp, or to ``super`` in Python.

Finally, let me notice that the decorator module implementation does
not use any cache, whereas the ``singledispatch`` implementation does.

Caveats and limitations
-------------------------------------------

In the present implementation, decorators generated by ``decorator``
can only be used on user-defined Python functions, methods or coroutines.
I have no interest in decorating generic callable objects. If you want to
decorate things like classmethods/staticmethods and general callables -
which I will never support in the decorator module - I suggest you
to look at the [wrapt](https://wrapt.readthedocs.io/en/latest/)
project by Graeme Dumpleton.

There is a strange quirk when decorating functions with keyword
arguments, if one of the arguments has the same name used in the
caller function for the first argument. The quirk was reported by
David Goldstein.

Here is an example where it is manifest:

```python
>>> @memoize
... def getkeys(**kw):
...     return kw.keys()

>>> getkeys(func='a') # doctest: +ELLIPSIS
Traceback (most recent call last):
 ...
TypeError: _memoize() got multiple values for ... 'func'

```

The error message looks really strange... until you realize that
the caller function `_memoize` uses `func` as first argument,
so there is a confusion between the positional argument and the
keyword arguments.

The solution is to change the name of the first argument in `_memoize`,
or to change the implementation like so:

```python

def _memoize(*all_args, **kw):
    func = all_args[0]
    args = all_args[1:]
    if kw:  # frozenset is used to ensure hashability
        key = args, frozenset(kw.items())
    else:
        key = args
    cache = func.cache  # attribute added by memoize
    if key not in cache:
        cache[key] = func(*args, **kw)
    return cache[key]
```

This avoids the need to name the first argument, so the problem
simply disappears. This is a technique that you should keep in mind
when writing decorators for functions with keyword arguments. Also,
notice that lately I have come to believe that decorating functions with
keyword arguments is not such a good idea, and you may want not to do
that.

The implementation is such that the decorated function makes
a (shallow) copy of the original function dictionary:

```python
>>> def f(): pass # the original function
>>> f.attr1 = "something" # setting an attribute
>>> f.attr2 = "something else" # setting another attribute

>>> traced_f = trace(f) # the decorated function

>>> traced_f.attr1
'something'
>>> traced_f.attr2 = "something different" # setting attr
>>> f.attr2 # the original attribute did not change
'something else'

```

Finally, you should be aware of the performance penalty of decorators.
The worse case is shown by the following example:

```bash
 $ cat performance.sh
 python3 -m timeit -s "
 from decorator import decorator

 @decorator
 def do_nothing(func, *args, **kw):
     return func(*args, **kw)

 @do_nothing
 def f():
     pass
 " "f()"

 python3 -m timeit -s "
 def f():
     pass
 " "f()"

```
On my laptop, using the ``do_nothing`` decorator instead of the
plain function is five times slower:

```bash
 $ bash performance.sh
 1000000 loops, best of 3: 1.39 usec per loop
 1000000 loops, best of 3: 0.278 usec per loop
```

Of course, a real life function probably does something more useful
than the function ``f`` here, so the real life performance penalty
*could* be negligible.  As always, the only way to know if there is a
penalty in your specific use case is to measure it.

LICENSE (2-clause BSD)
---------------------------------------------

Copyright (c) 2005-2020, Michele Simionato
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
  Redistributions in bytecode form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the
  distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

If you use this software and you are happy with it, consider sending me a
note, just to gratify my ego. On the other hand, if you use this software and
you are unhappy with it, send me a patch!
"""

function_annotations = """Function annotations
---------------------------------------------

Python 3 introduced the concept of [function annotations](
http://www.python.org/dev/peps/pep-3107/): the ability
to annotate the signature of a function with additional information,
stored in a dictionary named ``__annotations__``. The ``decorator`` module
(starting from release 3.3) will understand and preserve these annotations.

Here is an example:

```python
>>> @trace
... def f(x: 'the first argument', y: 'default argument'=1, z=2,
...       *args: 'varargs', **kw: 'kwargs'):
...     pass

```

In order to introspect functions with annotations, one needs
``inspect.getfullargspec`` (introduced in Python 3, then
deprecated in Python 3.5, then undeprecated in Python 3.6):

```python
>>> from inspect import getfullargspec
>>> argspec = getfullargspec(f)
>>> argspec.args
['x', 'y', 'z']
>>> argspec.varargs
'args'
>>> argspec.varkw
'kw'
>>> argspec.defaults
(1, 2)
>>> argspec.kwonlyargs
[]
>>> argspec.kwonlydefaults

```

You can check that the ``__annotations__`` dictionary is preserved:

```python
>>> f.__annotations__ is f.__wrapped__.__annotations__
True

```

Here ``f.__wrapped__`` is the original undecorated function.
This attribute exists for consistency with the behavior of
``functools.update_wrapper``.

Another attribute copied from the original function is ``__qualname__``,
the qualified name. This attribute was introduced in Python 3.3.
"""
if sys.version_info < (3,):
    function_annotations = ''

today = time.strftime('%Y-%m-%d')

__doc__ = (doc.replace('$VERSION', __version__).replace('$DATE', today)
           .replace('$FUNCTION_ANNOTATIONS', function_annotations))


def decorator_apply(dec, func):
    """
    Decorate a function by preserving the signature even if dec
    is not a signature-preserving decorator.
    """
    return FunctionMaker.create(
        func, 'return decfunc(%(signature)s)',
        dict(decfunc=dec(func)), __wrapped__=func)


def _trace(f, *args, **kw):
    kwstr = ', '.join('%r: %r' % (k, kw[k]) for k in sorted(kw))
    print("calling %s with args %s, {%s}" % (f.__name__, args, kwstr))
    return f(*args, **kw)


def trace(f):
    return decorate(f, _trace)


class Future(threading.Thread):
    """
    A class converting blocking functions into asynchronous
    functions by using threads.
    """
    def __init__(self, func, *args, **kw):
        try:
            counter = func.counter
        except AttributeError:  # instantiate the counter at the first call
            counter = func.counter = itertools.count(1)
        name = '%s-%s' % (func.__name__, next(counter))

        def func_wrapper():
            self._result = func(*args, **kw)
        super(Future, self).__init__(target=func_wrapper, name=name)
        self.start()

    def result(self):
        self.join()
        return self._result


def identity_dec(func):
    def wrapper(*args, **kw):
        return func(*args, **kw)
    return wrapper


@identity_dec
def example():
    pass


def memoize_uw(func):
    func.cache = {}

    def memoize(*args, **kw):
        if kw:  # frozenset is used to ensure hashability
            key = args, frozenset(kw.items())
        else:
            key = args
        if key not in func.cache:
            func.cache[key] = func(*args, **kw)
        return func.cache[key]
    return functools.update_wrapper(memoize, func)


@memoize_uw
def f1(x):
    "Simulate some long computation"
    time.sleep(1)
    return x


def _memoize(func, *args, **kw):
    if kw:  # frozenset is used to ensure hashability
        key = args, frozenset(kw.items())
    else:
        key = args
    cache = func.cache  # attribute added by memoize
    if key not in cache:
        cache[key] = func(*args, **kw)
    return cache[key]


def memoize(f):
    """
    A simple memoize implementation. It works by adding a .cache dictionary
    to the decorated function. The cache will grow indefinitely, so it is
    your responsibility to clear it, if needed.
    """
    f.cache = {}
    return decorate(f, _memoize)


@decorator
def blocking(f, msg='blocking', *args, **kw):
    if not hasattr(f, "thread"):  # no thread running
        def set_result():
            f.result = f(*args, **kw)
        f.thread = threading.Thread(None, set_result)
        f.thread.start()
        return msg
    elif f.thread.is_alive():
        return msg
    else:  # the thread is ended, return the stored result
        del f.thread
        return f.result


class User(object):
    "Will just be able to see a page"


class PowerUser(User):
    "Will be able to add new pages too"


class Admin(PowerUser):
    "Will be able to delete pages too"


class PermissionError(Exception):
    """
    >>> a = Action()
    >>> a.user = User()
    >>> a.view() # ok
    >>> a.insert() # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    PermissionError: User does not have the permission to run insert!
    """


@decorator
def restricted(func, user_class=User, *args, **kw):
    "Restrict access to a given class of users"
    self = args[0]
    if isinstance(self.user, user_class):
        return func(*args, **kw)
    else:
        raise PermissionError(
            '%s does not have the permission to run %s!'
            % (self.user, func.__name__))


class Action(object):
    @restricted(user_class=User)
    def view(self):
        "Any user can view objects"

    @restricted(user_class=PowerUser)
    def insert(self):
        "Only power users can insert objects"

    @restricted(user_class=Admin)
    def delete(self):
        "Only the admin can delete objects"


class TailRecursive(object):
    """
    tail_recursive decorator based on Kay Schluehr's recipe
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496691
    with improvements by me and George Sakkis.
    """

    def __init__(self, func):
        self.func = func
        self.firstcall = True
        self.CONTINUE = object()  # sentinel

    def __call__(self, *args, **kwd):
        CONTINUE = self.CONTINUE
        if self.firstcall:
            func = self.func
            self.firstcall = False
            try:
                while True:
                    result = func(*args, **kwd)
                    if result is CONTINUE:  # update arguments
                        args, kwd = self.argskwd
                    else:  # last call
                        return result
            finally:
                self.firstcall = True
        else:  # return the arguments of the tail call
            self.argskwd = args, kwd
            return CONTINUE


def tail_recursive(func):
    return decorator_apply(TailRecursive, func)


@tail_recursive
def factorial(n, acc=1):
    "The good old factorial"
    if n == 0:
        return acc
    return factorial(n-1, n*acc)


def fact(n):  # this is not tail-recursive
    if n == 0:
        return 1
    return n * fact(n-1)


def a_test_for_pylons():
    """
    In version 3.1.0 decorator(caller) returned a nameless partial
    object, thus breaking Pylons. That must not happen again.

    >>> decorator(_memoize).__name__
    '_memoize'

    Here is another bug of version 3.1.1 missing the docstring:

    >>> factorial.__doc__
    'The good old factorial'
    """


def test_kwonlydefaults():
    """
    >>> @trace
    ... def f(arg, defarg=1, *args, kwonly=2): pass
    ...
    >>> f.__kwdefaults__
    {'kwonly': 2}
    """


def test_kwonlyargs():
    """
    >>> @trace
    ... def func(a, b, *args, y=2, z=3, **kwargs):
    ...     return y, z
    ...
    >>> func('a', 'b', 'c', 'd', 'e', y='y', z='z', cat='dog')
    calling func with args ('a', 'b', 'c', 'd', 'e'), {'cat': 'dog', 'y': 'y', 'z': 'z'}
    ('y', 'z')
    """


def test_kwonly_no_args():
    """# this was broken with decorator 3.3.3
    >>> @trace
    ... def f(**kw): pass
    ...
    >>> f()
    calling f with args (), {}
    """


def test_kwonly_star_notation():
    """
    >>> @trace
    ... def f(*, a=1, **kw): pass
    ...
    >>> import inspect
    >>> inspect.getfullargspec(f)
    FullArgSpec(args=[], varargs=None, varkw='kw', defaults=None, kwonlyargs=['a'], kwonlydefaults={'a': 1}, annotations={})
    """

# #######################  multiple dispatch ############################ #


class XMLWriter(object):
    def __init__(self, **config):
        self.cfg = config

    @dispatch_on('obj')
    def write(self, obj):
        raise NotImplementedError(type(obj))


@XMLWriter.write.register(float)
def writefloat(self, obj):
    return '<float>%s</float>' % obj


class Rock(object):
    ordinal = 0


class Paper(object):
    ordinal = 1


class Scissors(object):
    ordinal = 2


class StrongRock(Rock):
    pass


@dispatch_on('a', 'b')
def win(a, b):
    if a.ordinal == b.ordinal:
        return 0
    elif a.ordinal > b.ordinal:
        return -win(b, a)
    raise NotImplementedError((type(a), type(b)))


@win.register(Rock, Paper)
def winRockPaper(a, b):
    return -1


@win.register(Rock, Scissors)
def winRockScissors(a, b):
    return 1


@win.register(Paper, Scissors)
def winPaperScissors(a, b):
    return -1


@win.register(StrongRock, Paper)
def winStrongRockPaper(a, b):
    return 0


class WithLength(object):
    def __len__(self):
        return 0


class SomeSet(collections.abc.Sized):
    # methods that make SomeSet set-like
    # not shown ...
    def __len__(self):
        return 0


@dispatch_on('obj')
def get_length(obj):
    raise NotImplementedError(type(obj))


@get_length.register(collections.abc.Sized)
def get_length_sized(obj):
    return len(obj)


@get_length.register(collections.abc.Set)
def get_length_set(obj):
    return 1


class C(object):
    "Registered as Sized and Iterable"


collections.abc.Sized.register(C)
collections.abc.Iterable.register(C)


def singledispatch_example1():
    singledispatch = dispatch_on('obj')

    @singledispatch
    def g(obj):
        raise NotImplementedError(type(g))

    @g.register(collections.abc.Sized)
    def g_sized(object):
        return "sized"

    @g.register(collections.abc.Iterable)
    def g_iterable(object):
        return "iterable"

    g(C())  # RuntimeError: Ambiguous dispatch: Iterable or Sized?


def singledispatch_example2():
    # adapted from functools.singledispatch test case
    singledispatch = dispatch_on('arg')

    class S(object):
        pass

    class V(c.Sized, S):
        def __len__(self):
            return 0

    @singledispatch
    def g(arg):
        return "base"

    @g.register(S)
    def g_s(arg):
        return "s"

    @g.register(c.Container)
    def g_container(arg):
        return "container"

    v = V()
    assert g(v) == "s"
    c.Container.register(V)  # add c.Container to the virtual mro of V
    assert g(v) == "s"  # since the virtual mro is V, Sized, S, Container
    return g, V


@decorator
def warn_slow(func, duration=0, *args, **kwargs):
    t0 = time.time()
    res = func(*args, **kwargs)
    dt = time.time() - t0
    if dt >= duration:
        print('%s is slow' % func.__name__)
    return res


@warn_slow()  # with parens
def operation1():
    """
    >>> operation1()
    operation1 is slow
    """
    time.sleep(.1)


@warn_slow  # without parens
def operation2():
    """
    >>> operation2()
    operation2 is slow
    """
    time.sleep(.1)


def chatty(func, *args, **kwargs):
    print(args, sorted(kwargs.items()))
    return func(*args, **kwargs)


@decorator(chatty)
def printsum(x=1, y=2):
    print(x + y)


@decorator(chatty, kwsyntax=True)
def printsum2(x=1, y=2):
    print(x + y)


def chattywrapper(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        print(args, kwargs)
        return func(*args, **kwargs)
    return functools.wraps(wrapper)

# #######################  changing the signature ########################## #


# see https://github.com/micheles/decorator/pull/85
def to_method(f):
    """
    Takes a function with signature (..., context) and returns a new
    function with signature (self, ...) to be used a a method in a
    class with a .context attribute.
    """
    sig = inspect.signature(f)
    params = list(sig.parameters.values())
    assert params[-1].name == 'context'
    self = inspect.Parameter('self', inspect.Parameter.POSITIONAL_OR_KEYWORD)
    params.insert(0, self)  # insert self
    del params[-1]  # remove context
    newsig = '%s%s' % (f.__name__, sig.replace(parameters=params))
    return FunctionMaker.create(
        newsig, 'context = self.context; return _func_%s' % sig,
        dict(_func_=f))


def foo(x, context=None):
    return x


def bar(x, y, context):
    return x + y


class Client:
    def __init__(self, context):
        self.context = context


def test_change_sig():
    """
    >>> Client.foo = to_method(foo)
    >>> Client.bar = to_method(bar)
    >>> c = Client(None)
    >>> assert c.foo(1) == 1
    >>> assert c.bar(1, 2) == 3
    """


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
import traceback
import warnings

from Aspidites._vendor._compat import IS_PY3, unicode, basestring

from .interface import describe_type, describe_value  # @UnusedImport # old interface

__all__ = [
    'indent',
    'deprecated',
    'raise_type_mismatch',
    'raise_wrapped',
    'raise_desc',
    'check_isinstance',
]


def indent(s, prefix, first=None): # pragma: no cover
    if not isinstance(s, basestring):
        s = u'{}'.format(s)

    assert isinstance(prefix, basestring)
    try:
        lines = s.split('\n')
    except UnicodeDecodeError:
        print(type(s)) # XXX
        print(s) # XXX
        lines = [s]
    if not lines:
        return u''

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = ' ' * (m - len(prefix)) + prefix
    first = ' ' * (m - len(first)) + first

    # differnet first prefix
    res = [u'%s%s' % (prefix, line.rstrip()) for line in lines]
    res[0] = u'%s%s' % (first, lines[0].rstrip())
    return '\n'.join(res)


def deprecated(func): # pragma: no cover
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    def new_func(*args, **kwargs):
        warnings.warn("Call to deprecated function %s." % func.__name__,
                      category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


def check_isinstance(ob, expected, **kwargs): # pragma: no cover
    if not isinstance(ob, expected):
        kwargs['object'] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs): # pragma: no cover
    """ Raises an exception concerning ob having the wrong type. """
    e = 'Object not of expected type:'
    e += '\n  expected: {}'.format(expected)
    e += '\n  obtained: {}'.format(type(ob))
    e += '\n' + indent(format_obs(kwargs), ' ')
    raise ValueError(e)


def format_dict_long(d, informal=False): # pragma: no cover
    """
        k: value
           kooked
        k: value
    """
    if not d:
        return '{}'

    maxlen = max([len(n) for n in d]) + 2

    def pad(pre):
        return ' ' * (maxlen - len(pre)) + pre

    res = ""
    order = sorted(d)
    for i, name in enumerate(order):
        value = d[name]
        prefix = pad('%s: ' % name)
        if i > 0:
            res += '\n'

        s = _get_str(value, informal)

        if len(s) > 512:
            s = s[:512] + ' [truncated]'
        res += indent(s, ' ', first=prefix)
    return res


def _get_str(x, informal): # pragma: no cover
    from contracts.interface import describe_value_multiline
    if informal:
        s = x.__str__()
    else:
        s = describe_value_multiline(x)
    return s


def format_list_long(l, informal=False): # pragma: no cover
    """
        - My 
          first
        - Second
    """
    res = ""
    for i, value in enumerate(l):
        prefix = '- '
        if i > 0:
            res += '\n'
        s = _get_str(value, informal)
        res += indent(s, ' ', first=prefix)
    return res


def format_obs(d, informal=False): # pragma: no cover
    """ Shows objects values and typed for the given dictionary """
    if not d:
        return d.__str__()

    maxlen = 0
    for name in d:
        maxlen = max(len(name), maxlen)

    def pad(pre):
        return ' ' * (maxlen - len(pre)) + pre

    res = ''

    S = sorted(d)
    for i, name in enumerate(S):
        value = d[name]
        prefix = pad('%s: ' % name)
        if i > 0:
            res += '\n'

        s = _get_str(value, informal)

        res += indent(s, ' ', first=prefix)

    return res


def raise_wrapped(etype, e, msg, compact=False, **kwargs): # pragma: no cover
    """ Raises an exception of type etype by wrapping
        another exception "e" with its backtrace and adding
        the objects in kwargs as formatted by format_obs.
        
        if compact = False, write the whole traceback, otherwise just str(e).
    
        exc = output of sys.exc_info()
    """

    if IS_PY3:
        msg += '\n' + indent(e, '│ ')
        e2 = etype(_format_exc(msg, **kwargs))
        # e2 = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)
        try:
            raise e2 from e
        finally:
            e2 = None
        # raise e2
    else:
        e2 = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)
        raise e2


def raise_wrapped_make(etype, e, msg, compact=False, **kwargs): # pragma: no cover
    """ Constructs the exception to be thrown by raise_wrapped() """
    assert isinstance(e, BaseException), type(e)
    check_isinstance(msg, unicode)
    s = msg
    if kwargs:
        s += '\n' + format_obs(kwargs)

    # import sys
    # if sys.version_info[0] >= 3:
    #     es = e.__str__()
    # else:
    if compact:
        es = e.__str__()
    else:  # pragma: no cover
        es = traceback.format_exc()  # only PY2

    s += '\n' + indent(es.strip(), '| ')

    return etype(s)


def _format_exc(msg, **kwargs): # pragma: no cover
    check_isinstance(msg, unicode)
    s = msg
    if kwargs:
        s += '\n' + format_obs(kwargs)
    return s


def raise_desc(etype, msg, args_first=False, **kwargs):  # pragma: no cover
    """
    
        Example:
            raise_desc(ValueError, "I don't know", a=a, b=b)
    """
    assert isinstance(msg, basestring), type(msg)
    s1 = msg
    if kwargs:
        s2 = format_obs(kwargs)
    else:
        s2 = ""

    if args_first:
        s = s2 + "\n" + s1
    else:
        s = s1 + "\n" + s2

    raise etype(s)


#
#
#
# def format_tb(tb, limit = None):
#     """A shorthand for 'format_list(extract_stack(f, limit))."""
#     return format_list(extract_tb(tb, limit))
#
# def extract_tb(tb, limit = None):
#     """Return list of up to limit pre-processed entries from traceback.
#
#     This is useful for alternate formatting of stack traces.  If
#     'limit' is omitted or None, all entries are extracted.  A
#     pre-processed stack trace entry is a quadruple (filename, line
#     number, function name, text) representing the information that is
#     usually printed for a stack trace.  The text is a string with
#     leading and trailing whitespace stripped; if the source is not
#     available it is None.
#     """
#     if limit is None:
#         if hasattr(sys, 'tracebacklimit'):
#             limit = sys.tracebacklimit
#     list = []
#     n = 0
#     while tb is not None and (limit is None or n < limit):
#         f = tb.tb_frame
#         lineno = tb.tb_lineno
#         co = f.f_code
#         filename = co.co_filename
#         name = co.co_name
#         linecache.checkcache(filename)
#         line = linecache.getline(filename, lineno, f.f_globals)
#         if line: line = line.strip()
#         else: line = None
#         list.append((filename, lineno, name, line))
#         tb = tb.tb_next
#         n = n+1
#     return list


# from decorator import decorator  # @UnresolvedImport

def ignore_typeerror(f):  # pragma: no cover
    """ Recasts TypeError as Exception; otherwise pyparsing gets confused. """

    def f2(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except TypeError as e:
            raise Exception(traceback.format_exc())

    return f2



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.
from pyparsing import (
    Combine,
    Empty,
    FollowedBy,
    Forward,
    Group,
    Keyword,
    Literal,
    OneOrMore,
    OnlyOnce,
    Optional,
    ParseElementEnhance,
    ParseException,
    Regex,
    Suppress,
    Word,
    alphanums,
    alphas,
    col,
    conditionAsParseAction,
    delimitedList,
    infixNotation,
    matchOnlyAtCol,
    nums,
    oneOf,
    opAssoc,
    quotedString,
    replaceWith,
    unicodeString,
)

from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.contracts.syntax import contract_expression

_contract_expression = contract_expression.copy()
_contract_expression.setParseAction(lambda tks: f"'{''.join((str(t) for t in tks))}'")

available_pragmas = [
    'cython.binding',
    'cython.boundscheck',
    'cython.wraparound',
    'cython.initializedcheck',
    'cython.nonecheck',
    'cython.overflowcheck',
    'cython.overflowcheck.fold',
    'cython.embedsignature',
    'cython.cdivision',
    'cython.cdivision_warnings',
    'cython.always_allow_keywords',
    'cython.c_api_binop_methods',
    'cython.profile',
    'cython.linetrace',
    'cython.infer_types',
    'cython.type_version_tag',
    'cython.unraisable_tracebacks',
    'cython.iterable_coroutine',
    'cython.emit_code_comments',
    'cython.cpp_locals',
    'cython.optimize.use_switch',
    'cython.optimize.unpack_method_calls',
    'cython.warn.undeclared',
    'cython.warn.unreachable',
    'cython.warn.maybe_uninitialized',
    'cython.warn.unused',
    'cython.warn.unused_arg',
    'cython.warn.unused_result',
    'cython.warn.multiple_declarators'
]


class IndentedBlock(ParseElementEnhance):
    """
    Expression to match one or more expressions at a given indentation level.
    Useful for parsing text where structure is implied by indentation (like Python source code).
    """

    def __init__(self, expr, recursive=True):
        super().__init__(expr, savelist=True)
        self._recursive = recursive
        self.skipWhitespace = False

    def parseImpl(self, instring, loc, doActions=True):
        # see if self.expr matches at the current location - if not it will raise an exception
        # and no further work is necessary
        self.expr.parseImpl(instring, loc, doActions)

        indent_col = col(loc, instring)
        peer_parse_action = matchOnlyAtCol(indent_col)
        peer_expr = FollowedBy(self.expr).addParseAction(peer_parse_action)
        inner_expr = Empty() + peer_expr.suppress() + self.expr

        if self._recursive:
            indent_parse_action = conditionAsParseAction(
                lambda s, l, t, relative_to_col=indent_col: col(l, s) > relative_to_col
            )
            indent_expr = FollowedBy(self.expr).addParseAction(indent_parse_action)
            inner_expr += Optional(indent_expr + self)

        return OneOrMore(inner_expr).parseImpl(instring, loc, doActions)


def cvt_bool(t):
    return t[0] == "True"


def cvt_int(t):
    return int(t[0])


def cvt_real(t):
    return float(t[0])


def cvt_tuple(t):
    return "(" + ", ".join(t.asList()) + ")"


def cvt_dict(t):
    t = t.asList()
    for i, v in enumerate(t):
        key, val = v
        if isinstance(key, str):  # string keys only
            t[i] = f"{key}: {val}"
        else:  # integer key
            t[i] = f"{key}: {val}"
    t = f'{", ".join(t)}'
    return f"pmap({{{t}}})"


def cvt_list(t):
    t = t.asList()
    for i, v in enumerate(t):
        if isinstance(v, str):  # string keys only
            t[i] = v
        else:  # integer key
            t[i] = int(v)
    t = f'{", ".join(t)}'
    return f"pvector([{t}])"


def cvt_set(t):
    t = t.asList()
    for i, v in enumerate(t):
        if isinstance(v, str):  # string keys only
            t[i] = v
        else:  # integer key
            t[i] = int(v)
    t = f'{", ".join(t)}'
    return f"pset({{{t}}})"


def cvt_contract_assign(t):
    t = swap_val_to_idx(list(t), ":", 1)
    t[2], t[4] = t[4], t[2]
    return " ".join((str(t) for t in t))


def cvt_contract_define(t):
    t[0], t[1] = t[1], t[0]
    t[1] = "'" + t[1] + "'"
    args = f"({', '.join(t[1:])})"
    t = t[0] + args
    return "".join(t)


# TODO: ADD cvtSimpleDefine


def swap_val_to_idx(lst: list, val, idx: int) -> list:
    val_idx = lst.index(val)
    if val_idx == idx:
        pass  # maybe error?
    lst[val_idx], lst[idx] = lst[idx], lst[val_idx]
    return lst


lparen, rparen, lbrack, rbrack, lbrace, rbrace, colon, comma = map(Suppress, "()[]{}:,")
unistr = unicodeString().setParseAction(lambda t: t[0][2:-1])
quoted_str = quotedString().setParseAction(lambda t: t[0])
bool_literal = oneOf("True False", asKeyword=True).setParseAction(cvt_bool)
null = Keyword("None").setParseAction(replaceWith(None))
integer = Word(nums).setParseAction(cvt_int)
real = Combine(Word(nums) + "." + Word(nums))
complex_ = Combine(real | integer + "+" + real | integer + "j")
list_str = Forward()
set_str = Forward()
dict_str = Forward()
tuple_str = Forward()
simple_assign = Forward()
# for_stmt = Forward()
pass_stmt = Forward()
suite = Forward()
rvalue = Forward()
stmt = Forward()
underscore = Literal("_")
func_call = Forward()
identifier = Word(alphas + "-_", alphanums + "-_").setName("VAR_ID")
operand = complex_ | real | integer | identifier | underscore

signop = oneOf("+ -")
multop = oneOf("* / %")
plusop = oneOf("+ -")
exponp = Literal("**")
factop = Literal("!")


def cvt_arith_expr(tks):
    expr = "".join((str(t) for t in tks))
    if "/" in expr:
        return "Maybe(SafeDiv, " + expr.replace("/", ", ") + ")"
    elif "%" in expr:
        return "Maybe(SafeMod, " + expr.replace("%", ", ") + ")"
    elif "**" in expr:
        return "Maybe(SafeExp, " + expr.replace("**", ", ") + ")"


arith_expr = Combine(
    infixNotation(
        operand,
        [
            (factop, 1, opAssoc.LEFT),
            (exponp, 2, opAssoc.RIGHT),
            (signop, 1, opAssoc.RIGHT),
            (multop, 2, opAssoc.LEFT),
            (plusop, 2, opAssoc.LEFT),
        ],
        lpar="(",
        rpar=")",
    )
).setParseAction(cvt_arith_expr)


comparisonop = oneOf("< <= > >= != ==")
comp_expr = infixNotation(
    arith_expr,
    [
        (comparisonop, 2, opAssoc.LEFT),
    ],
)

list_item = (
    real
    | arith_expr
    | integer
    | identifier
    | complex_
    | quoted_str
    | unistr
    | bool_literal
    | null
    | list_str
    | set_str
    | tuple_str
    | dict_str
    | comp_expr
)

eq = Literal("=")
respects = Keyword("->").setParseAction(lambda t: ":")
imposes = Keyword("<-").setParseAction(lambda t: "new_contract")


identifier = Word(alphas + "_", alphanums + "_")
contract_define = identifier + imposes + _contract_expression  #  ^ funcCall
contract_define.setParseAction(cvt_contract_define)
contract_respect = respects + _contract_expression
contract_assign = identifier + eq + list_item + contract_respect
contract_assign.setParseAction(cvt_contract_assign)

tuple_str <<= (
    lparen + Optional(delimitedList(list_item)) + Optional(comma) + rparen
).setParseAction(cvt_tuple)

list_str <<= (
    lbrack + Optional(delimitedList(list_item) + Optional(comma)) + rbrack
).setParseAction(cvt_list)

set_str <<= (
    lbrace + Optional(delimitedList(list_item) + Optional(comma)) + rbrace
).setParseAction(cvt_set)

dict_entry = Group(list_item + colon + list_item)
dict_str <<= (
    lbrace + Optional(delimitedList(dict_entry) + Optional(comma)) + rbrace
).setParseAction(cvt_dict)


private_def_decl = Literal("(").setParseAction(replaceWith("def "))
def_args = Optional(delimitedList(contract_assign, delim=";")).setParseAction(
    lambda t: ", ".join(t)
)
args_end = Group(Literal(")") + Literal(")")).setParseAction(replaceWith(") -> "))
def_args = Group("(" + def_args + args_end).setParseAction(lambda t: "".join(*t))

std_decor = "@contract()\n"


def cvt_pragma(tks):
    t: list = tks.asList()
    return '\n'.join(t) + '\n'

pragmas = Combine(
    Literal("#").setParseAction(replaceWith('@'))
    + oneOf(' '.join(available_pragmas))
    + '(' + oneOf('True False') + ')'
).setParseAction(cvt_pragma)
func_decl = Group(Optional(OneOrMore(pragmas)) +
    private_def_decl + identifier + def_args + _contract_expression
).setParseAction(lambda t: std_decor + "".join(*t) + ":")
comment_line = (
    Combine(Regex(r"`(?:[^`\n\r\\]|(?:``)|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*") + "`")
    .setParseAction(lambda t: t[0])
    .setParseAction(
        lambda s, loc, t: "# comment_line %s:"
        % (len([c for c in s[:loc] if c == "\n"]) + 1)
        + t[0]
    )
)
for_loop = Literal("<@>").setParseAction(replaceWith("for "))
return_none = Literal("<*>").setParseAction(replaceWith("return "))
yield_none = Literal("<^>").setParseAction(replaceWith("yield "))
return_value = return_none + rvalue
yield_value = yield_none + rvalue
ret_stmt = Group(return_value).setParseAction(lambda t: "".join(*t))
yield_stmt = Group(yield_value).setParseAction(lambda t: "".join(*t))
func_def = Group(func_decl + suite).setParseAction(
    lambda t: "\n    ".join(t[0]) + "\n\n"
)
pass_stmt = Keyword("pass").setParseAction(lambda t: str(*t))
suite << IndentedBlock(
    (
        comment_line
        | pass_stmt
        | ret_stmt
        | yield_stmt
        | func_call
        | func_def
        | contract_assign
    )
).setParseAction(lambda t: ("\n    ".join(t.asList())))
sep = ", "
lambda_def = Combine(Group("(" + arith_expr | comp_expr + ")"))
func_call = Group(
    identifier + "(" + Optional(delimitedList(rvalue)) + ")"
).setParseAction(
    lambda t: "Maybe" + t[0][1] + t[0][0] + sep + sep.join(t[0][2:-1]) + t[0][-1] + "()"
)  # if len(t[0]) != 3 else t[0][0] + '()')
clos_call = Group(
    identifier + "(" + Optional(delimitedList(rvalue)) + ")"
).setParseAction(
    lambda t: "Maybe" + t[0][1] + t[0][0] + sep + sep.join(t[0][2:-1]) + t[0][-1]
) + Suppress(
    Literal("...")
)


def cvt_for_stmt(toks):
    for t in toks:
        t[0], t[1] = t[1], t[0]
        t.insert(2, " in ")
        t.append(":")
    return "".join(*toks)


# for_stmt = Group(delimitedList(identifier) + for_loop + tuple_str | list_str |
#                  funcCall).setParseAction(cvt_for_stmt)
rvalue << (clos_call | func_call | list_item | lambda_def)
simple_assign << Group(identifier + "=" + rvalue).setParseAction(
    lambda t: " ".join(t[0])
)
stmt << (func_def | contract_define | func_call | simple_assign | comment_line)

_main = Keyword("main:").setParseAction(replaceWith('if __name__ == "__main__":')) + OneOrMore(stmt).setParseAction(lambda t: '    ' + '\n    '.join(t))

module_body = OneOrMore(stmt) + Optional(_main)

@new_contract
def woma_module(text: 'str'):
    try:
        module_body.parseString(text, parseAll=True)
    except ParseException as e:
        raise ValueError(str(e))
    else:
        return True

@contract
def parse_module(module: 'woma_module'):
    return module_body.parseString(module, parseAll=True)


# if __name__ == "__main__":
#     print(parse_module(open('../examples/math.wom').read()))



#!/usr/bin/python3
import os
from setuptools import setup, find_packages
from setuptools.dist import Distribution
from setuptools.command.install import install
from Aspidites import __version__, __license__, __title__, __author__, compiler, parser
from Aspidites.__main__ import get_cy_kwargs

cy_kwargs = get_cy_kwargs()
cy_kwargs.update({'embed': True})
code = open('Aspidites/woma/library.wom', 'r').read()
compiler.compile_module(
    parser.parse_module(code),
    'Aspidites/woma/library.pyx',
    bytecode=True,
    **cy_kwargs
)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


# Tested with wheel v0.29.0, 0.36.2
class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""
    def has_ext_modules(foo):
        return True


class InstallWrapper(install):
    """Provides a install wrapper for native woma
    extensions. These don't really belong in the
    Python package."""

    def run(self):
        # Run this first so the install stops in case
        # these fail otherwise the Python package is
        # successfully installed
        print("running preinstall hooks")
        self.preinstall()
        install.run(self)  # pip install
        print("running postinstall hooks")
        self.postinstall()

    def preinstall(self):
        """preinstall hook"""
        c = "Aspidites build/lib/Aspidites/woma/library.wom -c -o build/lib/Aspidites/woma/library.pyx --embed=True"
        os.popen(c)
        print(c)
        pass

    def postinstall(self):
        """postinstall hook"""
        pass


setup(
    name=__title__,
    version=__version__,
    author=__author__,
    author_email="rjdbcm@mail.umkc.edu",
    description="Aspidites is the reference implementation of the Woma Language",
    license=__license__,
    keywords="language",
    url="https://github.com/rjdbcm/Aspidites",
    install_requires=[
        'pyrsistent',
        'numpy',
        'cython>0.28,<3',
        'pyparsing',
        'mypy',
        'pytest',
        'future'
        ],
    packages=find_packages(),
    distclass=BinaryDistribution,
    entry_points={'console_scripts': ['aspidites = Aspidites.__main__:main']},
    package_data={'': ["*.wom", "*.pyx", "*.pyi"]},  # add any native *.wom files
    long_description=read('README.md'),
    cmdclass={'install': InstallWrapper},
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
)



from ..interface import Contract, ContractNotRespected, add_prefix
from ..pyparsing_utils import myOperatorPrecedence
from ..syntax import ParsingTmp, W, opAssoc, simple_contract
from .extensions import Extension
from .suggester import create_suggester

NOT_GLYPH = '!'
AND_GLYPH = ','
OR_GLYPH = '|'

class Logical(object):
    def __init__(self, glyph, precedence):
        self.glyph = glyph
        self.precedence = precedence

    def __str__(self):
        s = self.glyph.join(self._convert(x) for x in self.clauses)
        return s

    def _convert(self, x):
        if isinstance(x, Logical) and x.precedence < self.precedence:
            return '(%s)' % x
        return '%s' % x

class OR(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2
        Contract.__init__(self, where)
        Logical.__init__(self, OR_GLYPH, 1)
        self.clauses = clauses

    def _check_quick(self, context, value):
        """ Returns True if this checks out. """

        # first make a quick pass
        for c in self.clauses:
            try:
                # try with fake context
                c._check_contract(context.copy(), value, silent=True)
                # if ok, do with main context
                c._check_contract(context, value, silent=True)
                return True
            except ContractNotRespected as e:
                pass
        
        return False


    def check_contract(self, context, value, silent):
        orig = context.copy()

        if self._check_quick(context, value):
            return
        else:
            if silent:
                msg = '(Error description suppressed.)'
                raise ContractNotRespected(contract=self, error=msg,
                        value=value, context=context)

            # otherwise need to do it again with detailed error messages
            self.get_error(orig, value)

    def get_error(self, context, value):
        """ This assumes that we are going to fail """
        exceptions = []
        for c in self.clauses:
            try:
                # try with fake context
                c._check_contract(context.copy(), value, silent=False)
                # if ok, do with main context
                c._check_contract(context, value, silent=False)

                assert False, "We should not be here."
            except ContractNotRespected as e:
                exceptions.append((c, e))
        else:
            msg = self._format_exceptions(exceptions)
            raise ContractNotRespected(contract=self, error=msg,
                        value=value, context=context)

    def _format_exceptions(self, exceptions):
        msg = ('Could not satisfy any of the %d clauses in %s.'
               % (len(self.clauses), self))

        for i, ex in enumerate(exceptions):
            c, e = ex
            if i + 1 > 1:
                msg += '\n ├┄┄┄ Clause #%d:   %s\n' % (i + 1, c)
            else:
                msg += '\n ┬┄┄┄┄ Clause #%d:   %s\n' % (i + 1, c)
            msg += add_prefix('%s' % e, ' | ')

        msg += '\n ╰┄┄┄┄┄┄ (end clauses) ┄┄┄┄┄┄'
        return msg

    def __repr__(self):
        s = 'OR(%r)' % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == OR_GLYPH
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return OR(clauses, where=where)


class And(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2, clauses
        Contract.__init__(self, where)
        Logical.__init__(self, AND_GLYPH, 2)
        self.clauses = clauses

    def check_contract(self, context, value, silent):
        for c in self.clauses:
            c._check_contract(context, value, silent)

    def __repr__(self):
        s = 'And(%r)' % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == AND_GLYPH
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return And(clauses, where=where)


class Not(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) == 1, clauses
        Contract.__init__(self, where)
        Logical.__init__(self, NOT_GLYPH, 3)
        self.clauses = clauses

    def check_contract(self, context, value, silent):
        clause = self.clauses[0]
        try:
            clause._check_contract(context, value, silent)
        except ContractNotRespected:
            pass
        else:
            msg = "Shouldn't have satisfied the clause %s." % clause
            raise ContractNotRespected(contract=self, error=msg,
                                       value=value, context=context)

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        assert l.pop(0) == NOT_GLYPH
        where = W(string, location)
        return Not(l, where=where)

    def __repr__(self):
        s = 'Not(%r)' % self.clauses
        return s

    def __str__(self):
        return self.glyph + self._convert(self.clauses[0])


suggester = create_suggester(get_options=lambda: ParsingTmp.keywords +
                             list(Extension.registrar.keys()))
baseExpr = simple_contract | suggester
baseExpr.setName('Simple contract (recovering)')


op = myOperatorPrecedence
# op = operatorPrecedence
composite_contract = op(baseExpr, [
                         (NOT_GLYPH, 1, opAssoc.RIGHT, Not.parse_action),
                         (AND_GLYPH, 2, opAssoc.LEFT, And.parse_action),
                         (OR_GLYPH, 2, opAssoc.LEFT, OR.parse_action),
                    ])
composite_contract.setName('NOT/OR/AND contract')

or_contract = op(baseExpr, [
                         (OR_GLYPH, 2, opAssoc.LEFT, OR.parse_action),
                    ])
or_contract.setName('OR contract')
